import pytest

from supriya import default
from supriya.assets.synthdefs import test_two_voice
from supriya.contexts.nonrealtime import Score
from supriya.osc import OscBundle, OscMessage
from supriya.ugens import compile_synthdefs


@pytest.fixture
def context():
    return Score()


def test_add_group(context):
    with context.at(0):
        # /g_new
        group = context.add_group()
        # /p_new
        context.add_group(parallel=True)
    with context.at(1.23):
        context.add_group(parallel=True, target_node=group)
        context.add_group(parallel=True, target_node=group)
        context.add_group(target_node=group)
        context.add_group(target_node=group)
    assert list(context.iterate_osc_bundles()) == [
        OscBundle(
            contents=(
                OscMessage("/g_new", 1000, 0, 0),
                OscMessage("/p_new", 1001, 0, 0),
            ),
            timestamp=0.0,
        ),
        OscBundle(
            contents=(
                OscMessage("/p_new", 1002, 0, 1000, 1003, 0, 1000),
                OscMessage("/g_new", 1004, 0, 1000, 1005, 0, 1000),
            ),
            timestamp=1.23,
        ),
    ]
    # guard against invalid add-actions
    with context.at(2.34):
        synth = context.add_synth(default)
        with pytest.raises(ValueError):
            synth.add_group(add_action="ADD_TO_HEAD")
        with pytest.raises(ValueError):
            synth.add_group(add_action="ADD_TO_TAIL")


def test_add_synth(context):
    def compiled(*synthdefs):
        return compile_synthdefs(*synthdefs)

    with context.at(0):
        context.add_synthdefs(default)
        bus_a = context.add_bus("AUDIO")
        bus_c = context.add_bus("CONTROL")
        synth = context.add_synth(default)
        context.add_synth(default, frequency=bus_a, amplitude="c0", pan=0.25, out=0)
        context.add_synth(test_two_voice, frequencies=(123, 456))
    with context.at(1.23):
        context.add_synth(default, add_action="ADD_AFTER", target_node=synth)
        context.add_synth(default, frequency=bus_c, amplitude="a16", pan=0.25, out=0)
    assert list(context.iterate_osc_bundles()) == [
        OscBundle(
            contents=(
                OscMessage("/d_recv", compiled(default)),
                OscMessage("/s_new", "default", 1000, 0, 0),
                OscMessage(
                    "/s_new",
                    "default",
                    1001,
                    0,
                    0,
                    "amplitude",
                    "c0",
                    "frequency",
                    "a16",
                    "pan",
                    0.25,
                ),
                OscMessage(
                    "/s_new",
                    "test_two_voice",
                    1002,
                    0,
                    0,
                    "frequencies",
                    (123.0, 456.0),
                ),
            ),
            timestamp=0.0,
        ),
        OscBundle(
            contents=(
                OscMessage("/s_new", "default", 1003, 3, 1000),
                OscMessage(
                    "/s_new",
                    "default",
                    1004,
                    0,
                    0,
                    "amplitude",
                    "a16",
                    "frequency",
                    "c0",
                    "pan",
                    0.25,
                ),
            ),
            timestamp=1.23,
        ),
    ]
    # guard against invalid add-actions
    with context.at(2.34):
        synth = context.add_synth(default)
        with pytest.raises(ValueError):
            synth.add_synth(default, add_action="ADD_TO_HEAD")
        with pytest.raises(ValueError):
            synth.add_synth(default, add_action="ADD_TO_TAIL")


def test_free_group_children(context):
    with context.at(0):
        grandparent = context.add_group()
        # setup
        parent = grandparent.add_group()
        grandparent.add_synth(default)
        parent.add_synth(default)
        # /g_freeAll
        grandparent.free_children()
        # /g_deepFree
        grandparent.free_children(synths_only=True)
    assert list(context.iterate_osc_bundles()) == [
        OscBundle(
            contents=(
                OscMessage("/g_new", 1000, 0, 0, 1001, 0, 1000),
                OscMessage("/s_new", "default", 1002, 0, 1000),
                OscMessage("/s_new", "default", 1003, 0, 1001),
                OscMessage("/g_freeAll", 1000),
                OscMessage("/g_deepFree", 1000),
            ),
            timestamp=0.0,
        )
    ]


def test_free_node(context):
    with context.at(0):
        group = context.add_group()
        synth = context.add_synth(default)
        group.free()
        synth.free()
    assert list(context.iterate_osc_bundles()) == [
        OscBundle(
            contents=(
                OscMessage("/g_new", 1000, 0, 0),
                OscMessage("/s_new", "default", 1001, 0, 0),
                OscMessage("/n_free", 1000),
                OscMessage("/n_set", 1001, "gate", 0),
            ),
            timestamp=0.0,
        )
    ]


def test_map_node(context):
    with context.at(0):
        bus_a = context.add_bus("AUDIO")
        bus_c = context.add_bus("CONTROL")
        synth = context.add_synth(default)
        synth.map(frequency=bus_a, amplitude=bus_c, pan=None)
    assert list(context.iterate_osc_bundles()) == [
        OscBundle(
            contents=(
                OscMessage("/s_new", "default", 1000, 0, 0),
                OscMessage("/n_map", 1000, "amplitude", 0, "pan", -1),
                OscMessage("/n_mapa", 1000, "frequency", 16),
            ),
            timestamp=0.0,
        )
    ]


def test_move_node(context):
    with context.at(0):
        group = context.add_group()
        synth = context.add_synth(default)
        group.move(target_node=synth, add_action="ADD_AFTER")
        synth.move(target_node=group, add_action="ADD_BEFORE")
        synth.move(target_node=group, add_action="ADD_TO_TAIL")
        synth.move(target_node=group, add_action="ADD_TO_HEAD")
    assert list(context.iterate_osc_bundles()) == [
        OscBundle(
            contents=(
                OscMessage("/g_new", 1000, 0, 0),
                OscMessage("/s_new", "default", 1001, 0, 0),
                OscMessage("/n_after", 1000, 1001),
                OscMessage("/n_before", 1001, 1000),
                OscMessage("/g_tail", 1000, 1001),
                OscMessage("/g_head", 1000, 1001),
            ),
            timestamp=0.0,
        )
    ]


def test_order_nodes(context):
    with context.at(0):
        group_a = context.add_group()
        group_b = context.add_group()
        group_c = context.add_group()
        group_a.order(group_b, group_c, add_action="ADD_TO_TAIL")
    assert list(context.iterate_osc_bundles()) == [
        OscBundle(
            contents=(
                OscMessage("/g_new", 1000, 0, 0, 1001, 0, 0, 1002, 0, 0),
                OscMessage("/n_order", 1, 1000, 1001, 1002),
            ),
            timestamp=0.0,
        )
    ]


def test_pause_node(context):
    with context.at(0):
        group = context.add_group()
        group.pause()
    assert list(context.iterate_osc_bundles()) == [
        OscBundle(
            contents=(OscMessage("/g_new", 1000, 0, 0), OscMessage("/n_run", 1000, 0)),
            timestamp=0.0,
        )
    ]


def test_set_node(context):
    with context.at(0):
        group = context.add_group()
        group.set((1, 2.3), (2, [3.4, 4.5]), foo=3.145, bar=4.5, baz=[1.23, 4.56])
    assert list(context.iterate_osc_bundles()) == [
        OscBundle(
            contents=(
                OscMessage("/g_new", 1000, 0, 0),
                OscMessage(
                    "/n_set",
                    1000,
                    1,
                    2.3,
                    2,
                    [3.4, 4.5],
                    "bar",
                    4.5,
                    "baz",
                    [1.23, 4.56],
                    "foo",
                    3.145,
                ),
            ),
            timestamp=0.0,
        )
    ]


def test_set_node_range(context):
    with context.at(0):
        group = context.add_group()
        group.set_range((2, [3.4, 4.5]), baz=[1.23, 4.56])
    assert list(context.iterate_osc_bundles()) == [
        OscBundle(
            contents=(
                OscMessage("/g_new", 1000, 0, 0),
                OscMessage("/n_setn", 1000, 2, 2, 3.4, 4.5, "baz", 2, 1.23, 4.56),
            ),
            timestamp=0.0,
        )
    ]


def test_unpause_node(context):
    with context.at(0):
        group_a = context.add_group()
        group_b = context.add_group()
        group_c = context.add_group()
        group_a.unpause()
    with context.at(1.23):
        group_b.unpause()
        group_c.unpause()
    assert list(context.iterate_osc_bundles()) == [
        OscBundle(
            contents=(
                OscMessage("/g_new", 1000, 0, 0, 1001, 0, 0, 1002, 0, 0),
                OscMessage("/n_run", 1000, 1),
            ),
            timestamp=0.0,
        ),
        OscBundle(contents=(OscMessage("/n_run", 1001, 1, 1002, 1),), timestamp=1.23),
    ]



# flake8: noqa
import os
import platform

import pytest
from uqbar.strings import normalize

from supriya.ugens import (
    FFT,
    MFCC,
    In,
    Out,
    SuperColliderSynthDef,
    SynthDefBuilder,
    decompile_synthdef,
)


@pytest.fixture
def py_synthdef_mfcc():
    with SynthDefBuilder() as builder:
        source = In.ar(bus=0)
        pv_chain = FFT.kr(source=source)
        mfcc = MFCC.kr(pv_chain=pv_chain)
        Out.kr(bus=0, source=mfcc)
    return builder.build("MFCCTest")


@pytest.fixture
def sc_synthdef_mfcc():
    return SuperColliderSynthDef(
        "MFCCTest",
        r"""
        var in, fft, array;
        in = In.ar(0);
        fft = FFT(LocalBuf(2048), in);
        array = MFCC.kr(fft);
        Out.kr(0, array);
        """,
    )


@pytest.mark.skipif(platform.system() == "Windows", reason="hangs on Windows")
@pytest.mark.skipif(
    platform.system() == "Darwin" and os.environ.get("CI") == "true",
    reason="sclang hangs without QT",
)
def test_ugens(py_synthdef_mfcc, sc_synthdef_mfcc):
    py_ugens = tuple(repr(_) for _ in py_synthdef_mfcc.ugens)
    assert py_ugens == (
        "<In.ar()>",
        "<MaxLocalBufs.ir()>",
        "<LocalBuf.ir()>",
        "<FFT.kr()>",
        "<MFCC.kr()>",
        "<Out.kr()>",
    )
    sc_compiled_synthdef = bytes(sc_synthdef_mfcc.compile())
    sc_synthdef = decompile_synthdef(sc_compiled_synthdef)
    sc_ugens = tuple(repr(_) for _ in sc_synthdef.ugens)
    assert py_ugens == sc_ugens


@pytest.mark.skipif(platform.system() == "Windows", reason="hangs on Windows")
@pytest.mark.skipif(
    platform.system() == "Darwin" and os.environ.get("CI") == "true",
    reason="sclang hangs without QT",
)
def test_format(py_synthdef_mfcc, sc_synthdef_mfcc):
    py_format = str(py_synthdef_mfcc)
    assert py_format == normalize(
        """
        synthdef:
            name: MFCCTest
            ugens:
            -   In.ar:
                    bus: 0.0
            -   MaxLocalBufs.ir:
                    maximum: 1.0
            -   LocalBuf.ir:
                    channel_count: 1.0
                    frame_count: 2048.0
            -   FFT.kr:
                    buffer_id: LocalBuf.ir[0]
                    source: In.ar[0]
                    hop: 0.5
                    window_type: 0.0
                    active: 1.0
                    window_size: 0.0
            -   MFCC.kr:
                    pv_chain: FFT.kr[0]
                    coeff_count: 13.0
            -   Out.kr:
                    bus: 0.0
                    source[0]: MFCC.kr[0]
                    source[1]: MFCC.kr[1]
                    source[2]: MFCC.kr[2]
                    source[3]: MFCC.kr[3]
                    source[4]: MFCC.kr[4]
                    source[5]: MFCC.kr[5]
                    source[6]: MFCC.kr[6]
                    source[7]: MFCC.kr[7]
                    source[8]: MFCC.kr[8]
                    source[9]: MFCC.kr[9]
                    source[10]: MFCC.kr[10]
                    source[11]: MFCC.kr[11]
                    source[12]: MFCC.kr[12]
        """
    )
    sc_compiled_synthdef = bytes(sc_synthdef_mfcc.compile())
    sc_synthdef = decompile_synthdef(sc_compiled_synthdef)
    sc_format = str(sc_synthdef)
    assert py_format == sc_format


@pytest.mark.skipif(platform.system() == "Windows", reason="hangs on Windows")
@pytest.mark.skipif(
    platform.system() == "Darwin" and os.environ.get("CI") == "true",
    reason="sclang hangs without QT",
)
def test_py_compile(py_synthdef_mfcc, sc_synthdef_mfcc):
    py_compiled_synthdef = py_synthdef_mfcc.compile()
    # fmt: off
    assert py_compiled_synthdef == (
        b"SCgf"
        b"\x00\x00\x00\x02"
        b"\x00\x01"
            b"\x08MFCCTest"
                b"\x00\x00\x00\x05"
                    b"\x00\x00\x00\x00"
                    b"?\x80\x00\x00"
                    b"E\x00\x00\x00"
                    b"?\x00\x00\x00"
                    b"AP\x00\x00"
                b"\x00\x00\x00\x00"
                b"\x00\x00\x00\x00"
                b"\x00\x00\x00\x06"
                    b"\x02In"
                        b"\x02"
                        b"\x00\x00\x00\x01\x00\x00\x00\x01\x00\x00\xff\xff\xff\xff\x00\x00\x00\x00\x02"
                    b"\x0cMaxLocalBufs"
                        b"\x00"
                        b"\x00\x00\x00\x01\x00\x00\x00\x01\x00\x00\xff\xff\xff\xff\x00\x00\x00\x01\x00"
                    b"\x08LocalBuf"
                        b"\x00"
                        b"\x00\x00\x00\x03\x00\x00\x00\x01\x00\x00\xff\xff\xff\xff\x00\x00\x00\x01\xff\xff\xff\xff\x00\x00\x00\x02\x00\x00\x00\x01\x00\x00\x00\x00\x00"
                    b"\x03FFT"
                        b"\x01"
                        b"\x00\x00\x00\x06\x00\x00\x00\x01\x00\x00\x00\x00\x00\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xff\xff\xff\xff\x00\x00\x00\x03\xff\xff\xff\xff\x00\x00\x00\x00\xff\xff\xff\xff\x00\x00\x00\x01\xff\xff\xff\xff\x00\x00\x00\x00\x01"
                    b"\x04MFCC"
                        b"\x01"
                        b"\x00\x00\x00\x02"
                        b"\x00\x00\x00\r"
                        b"\x00\x00"
                            b"\x00\x00\x00\x03"
                            b"\x00\x00\x00\x00"
                            b"\xff\xff\xff\xff"
                            b"\x00\x00\x00\x04"
                            b"\x01\x01\x01\x01"
                            b"\x01\x01\x01\x01"
                            b"\x01\x01\x01\x01"
                            b"\x01"
                    b"\x03Out"
                        b"\x01"
                        b"\x00\x00\x00\x0e\x00\x00\x00\x00\x00\x00\xff\xff\xff\xff\x00\x00\x00\x00\x00\x00\x00\x04\x00\x00\x00\x00\x00\x00\x00\x04\x00\x00\x00\x01\x00\x00\x00\x04\x00\x00\x00\x02\x00\x00\x00\x04\x00\x00\x00\x03\x00\x00\x00\x04\x00\x00\x00\x04\x00\x00\x00\x04\x00\x00\x00\x05\x00\x00\x00\x04\x00\x00\x00\x06\x00\x00\x00\x04\x00\x00\x00\x07\x00\x00\x00\x04\x00\x00\x00\x08\x00\x00\x00\x04\x00\x00\x00\t\x00\x00\x00\x04\x00\x00\x00\n\x00\x00\x00\x04\x00\x00\x00\x0b\x00\x00\x00\x04\x00\x00\x00\x0c"
            b"\x00\x00"
    )
    # fmt: on
    sc_compiled_synthdef = bytes(sc_synthdef_mfcc.compile())
    assert sc_compiled_synthdef == py_compiled_synthdef



import asyncio
import logging

import pytest
import pytest_asyncio

from supriya.contexts.errors import MomentClosed
from supriya.contexts.realtime import AsyncServer, Server
from supriya.osc import OscBundle, OscMessage
from supriya.ugens import Out, SinOsc, SynthDefBuilder, compile_synthdefs


async def get(x):
    if asyncio.iscoroutine(x):
        return await x
    return x


@pytest.fixture(autouse=True)
def use_caplog(caplog):
    caplog.set_level(logging.INFO)


@pytest_asyncio.fixture(autouse=True, params=[AsyncServer, Server])
async def context(request):
    context = request.param()
    await get(context.boot())
    yield context


@pytest.fixture
def synthdefs():
    with SynthDefBuilder(bus=0, frequency=440) as builder:
        Out.ar(bus=builder["bus"], source=SinOsc.ar(frequency=builder["frequency"]))
    synthdef_a = builder.build(name="synthdef-a")
    synthdef_b = builder.build(name="synthdef-b")
    synthdef_c = builder.build(name="synthdef-c")
    return [synthdef_a, synthdef_b, synthdef_c]


@pytest.mark.asyncio
async def test_add_synthdefs(context, synthdefs):
    def compiled(*x):
        return compile_synthdefs(*x)

    with context.osc_protocol.capture() as transcript:
        # no synthdefs provided
        with pytest.raises(ValueError):
            context.add_synthdefs()
        # /d_recv
        context.add_synthdefs(synthdefs[0])
        # multiples
        context.add_synthdefs(*synthdefs)
        # completion without moment via on_completion lambda succeeds
        context.add_synthdefs(synthdefs[1], on_completion=lambda ctx: ctx.add_group())
        # completion without moment errors
        with pytest.raises(MomentClosed):
            with context.add_synthdefs(synthdefs[2]):
                context.add_group()
        # completion inside moment succeeds
        with context.at(1.23):
            with context.add_synthdefs(synthdefs[2]):
                context.add_group()
    assert transcript.filtered(received=False, status=False) == [
        OscMessage("/d_recv", compiled(synthdefs[0])),
        OscMessage("/d_recv", compiled(*synthdefs)),
        OscMessage("/d_recv", compiled(synthdefs[1]), OscMessage("/g_new", 1000, 0, 1)),
        OscMessage("/d_recv", compiled(synthdefs[2])),
        OscBundle(
            contents=[
                OscMessage(
                    "/d_recv",
                    compiled(synthdefs[2]),
                    OscMessage("/g_new", 1001, 0, 1),
                )
            ],
            timestamp=1.23 + context.latency,
        ),
    ]


@pytest.mark.asyncio
async def test_free_synthdefs(context, synthdefs):
    with context.osc_protocol.capture() as transcript:
        # no synthdefs provided
        with pytest.raises(ValueError):
            context.free_synthdefs()
        # /d_free
        context.free_synthdefs(*synthdefs)
    assert transcript.filtered(received=False, status=False) == [
        OscMessage("/d_free", "synthdef-a", "synthdef-b", "synthdef-c")
    ]


@pytest.mark.asyncio
async def test_free_all_synthdefs(context):
    with context.osc_protocol.capture() as transcript:
        context.free_all_synthdefs()
    assert transcript.filtered(received=False, status=False) == [
        OscMessage("/d_freeAll")
    ]


@pytest.mark.asyncio
async def test_load_synthdefs(context, tmp_path):
    with context.osc_protocol.capture() as transcript:
        context.load_synthdefs(tmp_path / "a.scsyndef")
        # completion without moment via on_completion lambda succeeds
        context.load_synthdefs(
            tmp_path / "b.scsyndef", on_completion=lambda ctx: ctx.add_group()
        )
        # completion without moment errors
        with pytest.raises(MomentClosed):
            with context.load_synthdefs(tmp_path / "c.scsyndef"):
                context.add_group()
        # completion inside moment succeeds
        with context.at(1.23):
            with context.load_synthdefs(tmp_path / "c.scsyndef"):
                context.add_group()
    assert transcript.filtered(received=False, status=False) == [
        OscMessage("/d_load", str(tmp_path / "a.scsyndef")),
        OscMessage(
            "/d_load", str(tmp_path / "b.scsyndef"), OscMessage("/g_new", 1000, 0, 1)
        ),
        OscMessage("/d_load", str(tmp_path / "c.scsyndef")),
        OscBundle(
            contents=(
                OscMessage(
                    "/d_load",
                    str(tmp_path / "c.scsyndef"),
                    OscMessage("/g_new", 1001, 0, 1),
                ),
            ),
            timestamp=1.23 + context.latency,
        ),
    ]


@pytest.mark.asyncio
async def test_load_synthdefs_directory(context, tmp_path):
    with context.osc_protocol.capture() as transcript:
        context.load_synthdefs_directory(tmp_path / "a")
        # completion without moment via on_completion lambda succeeds
        context.load_synthdefs_directory(
            tmp_path / "b", on_completion=lambda ctx: ctx.add_group()
        )
        # completion without moment errors
        with pytest.raises(MomentClosed):
            with context.load_synthdefs_directory(tmp_path / "c"):
                context.add_group()
        # completion inside moment succeeds
        with context.at(1.23):
            with context.load_synthdefs_directory(tmp_path / "c"):
                context.add_group()
    assert transcript.filtered(received=False, status=False) == [
        OscMessage("/d_loadDir", str(tmp_path / "a")),
        OscMessage("/d_loadDir", str(tmp_path / "b"), OscMessage("/g_new", 1000, 0, 1)),
        OscMessage("/d_loadDir", str(tmp_path / "c")),
        OscBundle(
            contents=(
                OscMessage(
                    "/d_loadDir", str(tmp_path / "c"), OscMessage("/g_new", 1001, 0, 1)
                ),
            ),
            timestamp=1.23 + context.latency,
        ),
    ]



from typing import Any, Dict, Tuple

from ..enums import CalculationRate
from ..typing import Default
from .bufio import LocalBuf
from .core import OutputProxy, UGen, UGenOperable, param, ugen
from .info import BufFrames


@ugen(is_width_first=True)
class PV_ChainUGen(UGen):
    """
    Abstract base class for all phase-vocoder-chain unit generators.
    """

    @property
    def fft_size(self) -> UGenOperable:
        """
        Gets FFT size as UGen input.

        Returns ugen input.
        """
        input_ = self.inputs[0]
        assert isinstance(input_, OutputProxy)
        assert isinstance(input_.ugen, PV_ChainUGen)
        return input_.ugen.fft_size


@ugen(kr=True, is_width_first=True)
class FFT(PV_ChainUGen):
    """
    A fast Fourier transform.

    ::

        >>> buffer_id = supriya.ugens.LocalBuf.ir(frame_count=2048)
        >>> source = supriya.ugens.In.ar(bus=0)
        >>> fft = supriya.ugens.FFT.kr(
        ...     active=1,
        ...     buffer_id=buffer_id,
        ...     hop=0.5,
        ...     source=source,
        ...     window_size=0,
        ...     window_type=0,
        ... )
        >>> fft
        <FFT.kr()[0]>
    """

    ### CLASS VARIABLES ###

    buffer_id = param(Default())
    source = param()
    hop = param(0.5)
    window_type = param(0)
    active = param(1)
    window_size = param(0)

    ### PRIVATE METHODS ###

    def _postprocess_kwargs(
        self, calculation_rate: CalculationRate, **kwargs
    ) -> Tuple[CalculationRate, Dict[str, Any]]:
        if isinstance(kwargs["buffer_id"], Default):
            kwargs["buffer_id"] = LocalBuf.ir(frame_count=kwargs["window_size"] or 2048)
        return calculation_rate, kwargs

    ### PUBLIC PROPERTIES ###

    @property
    def fft_size(self) -> UGenOperable:
        """
        Gets FFT size as UGen input.

        Returns ugen input.
        """
        return BufFrames.ir(buffer_id=self.buffer_id)


@ugen(ar=True, kr=True, is_width_first=True)
class IFFT(UGen):
    """
    An inverse fast Fourier transform.

    ::

        >>> pv_chain = supriya.ugens.LocalBuf.ir(frame_count=2048)
        >>> ifft = supriya.ugens.IFFT.ar(
        ...     pv_chain=pv_chain,
        ...     window_size=0,
        ...     window_type=0,
        ... )
        >>> ifft
        <IFFT.ar()[0]>
    """

    pv_chain = param()
    window_type = param(0)
    window_size = param(0)


@ugen(kr=True, is_width_first=True)
class PV_Add(PV_ChainUGen):
    """
    Complex addition.

    ::

        >>> pv_chain_a = supriya.ugens.FFT.kr(
        ...     source=supriya.ugens.WhiteNoise.ar(),
        ... )
        >>> pv_chain_b = supriya.ugens.FFT.kr(
        ...     source=supriya.ugens.LFSaw.ar(),
        ... )
        >>> pv_add = supriya.ugens.PV_Add.kr(
        ...     pv_chain_a=pv_chain_a,
        ...     pv_chain_b=pv_chain_b,
        ... )
        >>> pv_add
        <PV_Add.kr()[0]>
    """

    pv_chain_a = param()
    pv_chain_b = param()


@ugen(kr=True, is_width_first=True)
class PV_BinScramble(PV_ChainUGen):
    """
    Scrambles bins.

    ::

        >>> pv_chain = supriya.ugens.FFT.kr(
        ...     source=supriya.ugens.WhiteNoise.ar(),
        ... )
        >>> pv_bin_scramble = supriya.ugens.PV_BinScramble.kr(
        ...     pv_chain=pv_chain,
        ...     trigger=0,
        ...     width=0.2,
        ...     wipe=0,
        ... )
        >>> pv_bin_scramble
        <PV_BinScramble.kr()[0]>
    """

    pv_chain = param()
    wipe = param(0)
    width = param(0.2)
    trigger = param(0)


@ugen(kr=True, is_width_first=True)
class PV_BinShift(PV_ChainUGen):
    """
    Shifts and stretches bin positions.

    ::

        >>> pv_chain = supriya.ugens.FFT.kr(
        ...     source=supriya.ugens.WhiteNoise.ar(),
        ... )
        >>> pv_bin_shift = supriya.ugens.PV_BinShift.kr(
        ...     pv_chain=pv_chain,
        ...     interpolate=0,
        ...     shift=0,
        ...     stretch=1,
        ... )
        >>> pv_bin_shift
        <PV_BinShift.kr()[0]>
    """

    pv_chain = param()
    stretch = param(1.0)
    shift = param(0.0)
    interpolate = param(0)


@ugen(kr=True, is_width_first=True)
class PV_BinWipe(PV_ChainUGen):
    """
    Copies low bins from one input and the high bins of the other.

    ::

        >>> pv_chain_a = supriya.ugens.FFT.kr(
        ...     source=supriya.ugens.WhiteNoise.ar(),
        ... )
        >>> pv_chain_b = supriya.ugens.FFT.kr(
        ...     source=supriya.ugens.LFSaw.ar(),
        ... )
        >>> pv_bin_wipe = supriya.ugens.PV_BinWipe.kr(
        ...     pv_chain_a=pv_chain_a,
        ...     pv_chain_b=pv_chain_b,
        ...     wipe=0,
        ... )
        >>> pv_bin_wipe
        <PV_BinWipe.kr()[0]>
    """

    pv_chain_a = param()
    pv_chain_b = param()
    wipe = param(0)


@ugen(kr=True, is_width_first=True)
class PV_BrickWall(PV_ChainUGen):
    """
    Zeros bins.

    - If wipe == 0 then there is no effect.
    - If wipe > 0 then it acts like a high pass filter, clearing bins from the bottom
      up.
    - If wipe < 0 then it acts like a low pass filter, clearing bins from the top down.

    ::

        >>> pv_chain = supriya.ugens.FFT.kr(
        ...     source=supriya.ugens.WhiteNoise.ar(),
        ... )
        >>> pv_brick_wall = supriya.ugens.PV_BrickWall.kr(
        ...     pv_chain=pv_chain,
        ...     wipe=0,
        ... )
        >>> pv_brick_wall
        <PV_BrickWall.kr()[0]>
    """

    pv_chain = param()
    wipe = param(0)


@ugen(kr=True, is_width_first=True)
class PV_ConformalMap(PV_ChainUGen):
    """
    Complex plane attack.

    ::

        >>> pv_chain = supriya.ugens.FFT.kr(
        ...     source=supriya.ugens.WhiteNoise.ar(),
        ... )
        >>> pv_conformal_map = supriya.ugens.PV_ConformalMap.kr(
        ...     aimag=0,
        ...     areal=0,
        ...     pv_chain=pv_chain,
        ... )
        >>> pv_conformal_map
        <PV_ConformalMap.kr()[0]>
    """

    pv_chain = param()
    areal = param(0)
    aimag = param(0)


@ugen(kr=True, is_width_first=True)
class PV_Conj(PV_ChainUGen):
    """
    Complex conjugate.

    ::

        >>> pv_chain = supriya.ugens.FFT.kr(
        ...     source=supriya.ugens.WhiteNoise.ar(),
        ... )
        >>> pv_conj = supriya.ugens.PV_Conj.kr(
        ...     pv_chain=pv_chain,
        ... )
        >>> pv_conj
        <PV_Conj.kr()[0]>
    """

    pv_chain = param()


@ugen(kr=True, is_width_first=True)
class PV_Copy(PV_ChainUGen):
    """
    Copies an FFT buffer.

    ::

        >>> pv_chain_a = supriya.ugens.FFT.kr(
        ...     source=supriya.ugens.WhiteNoise.ar(),
        ... )
        >>> pv_chain_b = supriya.ugens.FFT.kr(
        ...     source=supriya.ugens.LFSaw.ar(),
        ... )
        >>> pv_copy = supriya.ugens.PV_Copy.kr(
        ...     pv_chain_a=pv_chain_a,
        ...     pv_chain_b=pv_chain_b,
        ... )
        >>> pv_copy
        <PV_Copy.kr()[0]>
    """

    pv_chain_a = param()
    pv_chain_b = param()


@ugen(kr=True, is_width_first=True)
class PV_CopyPhase(PV_ChainUGen):
    """
    Copies magnitudes and phases.

    ::

        >>> pv_chain_a = supriya.ugens.FFT.kr(
        ...     source=supriya.ugens.WhiteNoise.ar(),
        ... )
        >>> pv_chain_b = supriya.ugens.FFT.kr(
        ...     source=supriya.ugens.LFSaw.ar(),
        ... )
        >>> pv_copy_phase = supriya.ugens.PV_CopyPhase.kr(
        ...     pv_chain_a=pv_chain_a,
        ...     pv_chain_b=pv_chain_b,
        ... )
        >>> pv_copy_phase
        <PV_CopyPhase.kr()[0]>
    """

    pv_chain_a = param()
    pv_chain_b = param()


@ugen(kr=True, is_width_first=True)
class PV_Diffuser(PV_ChainUGen):
    """
    Shifts phases randomly.

    ::

        >>> pv_chain = supriya.ugens.FFT.kr(
        ...     source=supriya.ugens.WhiteNoise.ar(),
        ... )
        >>> pv_diffuser = supriya.ugens.PV_Diffuser.kr(
        ...     pv_chain=pv_chain,
        ...     trigger=0,
        ... )
        >>> pv_diffuser
        <PV_Diffuser.kr()[0]>
    """

    pv_chain = param()
    trigger = param(0)


@ugen(kr=True, is_width_first=True)
class PV_Div(PV_ChainUGen):
    """
    Complex division.

    ::

        >>> pv_chain_a = supriya.ugens.FFT.kr(
        ...     source=supriya.ugens.WhiteNoise.ar(),
        ... )
        >>> pv_chain_b = supriya.ugens.FFT.kr(
        ...     source=supriya.ugens.LFSaw.ar(),
        ... )
        >>> pv_div = supriya.ugens.PV_Div.kr(
        ...     pv_chain_a=pv_chain_a,
        ...     pv_chain_b=pv_chain_b,
        ... )
        >>> pv_div
        <PV_Div.kr()[0]>
    """

    pv_chain_a = param()
    pv_chain_b = param()


@ugen(kr=True, is_width_first=True)
class PV_HainsworthFoote(PV_ChainUGen):
    """
    A FFT onset detector.

    ::

        >>> pv_chain = supriya.ugens.FFT.kr(
        ...     source=supriya.ugens.WhiteNoise.ar(),
        ... )
        >>> pv_hainsworth_foote = supriya.ugens.PV_HainsworthFoote.kr(
        ...     pv_chain=pv_chain,
        ...     propf=0,
        ...     proph=0,
        ...     threshold=1,
        ...     waittime=0.04,
        ... )
        >>> pv_hainsworth_foote
        <PV_HainsworthFoote.kr()[0]>
    """

    pv_chain = param()
    proph = param(0)
    propf = param(0)
    threshold = param(1)
    waittime = param(0.04)


@ugen(kr=True, is_width_first=True)
class PV_JensenAndersen(PV_ChainUGen):
    """
    A FFT feature detector for onset detection.

    ::

        >>> pv_chain = supriya.ugens.FFT.kr(
        ...     source=supriya.ugens.WhiteNoise.ar(),
        ... )
        >>> pv_jensen_andersen = supriya.ugens.PV_JensenAndersen.kr(
        ...     pv_chain=pv_chain,
        ...     prophfc=0.25,
        ...     prophfe=0.25,
        ...     propsc=0.25,
        ...     propsf=0.25,
        ...     threshold=1,
        ...     waittime=0.04,
        ... )
        >>> pv_jensen_andersen
        <PV_JensenAndersen.kr()[0]>
    """

    pv_chain = param()
    propsc = param(0.25)
    prophfe = param(0.25)
    prophfc = param(0.25)
    propsf = param(0.25)
    threshold = param(1)
    waittime = param(0.04)


@ugen(kr=True, is_width_first=True)
class PV_LocalMax(PV_ChainUGen):
    """
    Passes bins which are local maxima.

    ::

        >>> pv_chain = supriya.ugens.FFT.kr(
        ...     source=supriya.ugens.WhiteNoise.ar(),
        ... )
        >>> pv_local_max = supriya.ugens.PV_LocalMax.kr(
        ...     pv_chain=pv_chain,
        ...     threshold=0,
        ... )
        >>> pv_local_max
        <PV_LocalMax.kr()[0]>
    """

    pv_chain = param()
    threshold = param(0)


@ugen(kr=True, is_width_first=True)
class PV_MagAbove(PV_ChainUGen):
    """
    Passes magnitudes above threshold.

    ::

        >>> pv_chain = supriya.ugens.FFT.kr(
        ...     source=supriya.ugens.WhiteNoise.ar(),
        ... )
        >>> pv_mag_above = supriya.ugens.PV_MagAbove.kr(
        ...     pv_chain=pv_chain,
        ...     threshold=0,
        ... )
        >>> pv_mag_above
        <PV_MagAbove.kr()[0]>
    """

    pv_chain = param()
    threshold = param(0)


@ugen(kr=True, is_width_first=True)
class PV_MagBelow(PV_ChainUGen):
    """
    Passes magnitudes below threshold.

    ::

        >>> pv_chain = supriya.ugens.FFT.kr(
        ...     source=supriya.ugens.WhiteNoise.ar(),
        ... )
        >>> pv_mag_below = supriya.ugens.PV_MagBelow.kr(
        ...     pv_chain=pv_chain,
        ...     threshold=0,
        ... )
        >>> pv_mag_below
        <PV_MagBelow.kr()[0]>
    """

    pv_chain = param()
    threshold = param(0)


@ugen(kr=True, is_width_first=True)
class PV_MagClip(PV_ChainUGen):
    """
    Clips magnitudes.

    ::

        >>> pv_chain = supriya.ugens.FFT.kr(
        ...     source=supriya.ugens.WhiteNoise.ar(),
        ... )
        >>> pv_mag_clip = supriya.ugens.PV_MagClip.kr(
        ...     pv_chain=pv_chain,
        ...     threshold=0,
        ... )
        >>> pv_mag_clip
        <PV_MagClip.kr()[0]>
    """

    pv_chain = param()
    threshold = param(0)


@ugen(kr=True, is_width_first=True)
class PV_MagDiv(PV_ChainUGen):
    """
    Divides magnitudes.

    ::

        >>> pv_chain_a = supriya.ugens.FFT.kr(
        ...     source=supriya.ugens.WhiteNoise.ar(),
        ... )
        >>> pv_chain_b = supriya.ugens.FFT.kr(
        ...     source=supriya.ugens.LFSaw.ar(),
        ... )
        >>> pv_mag_div = supriya.ugens.PV_MagDiv.kr(
        ...     pv_chain_a=pv_chain_a,
        ...     pv_chain_b=pv_chain_b,
        ...     zeroed=0.0001,
        ... )
        >>> pv_mag_div
        <PV_MagDiv.kr()[0]>
    """

    pv_chain_a = param()
    pv_chain_b = param()
    zeroed = param(0.0001)


@ugen(kr=True, is_width_first=True)
class PV_MagFreeze(PV_ChainUGen):
    """
    Freezes magnitudes.

    ::

        >>> pv_chain = supriya.ugens.FFT.kr(
        ...     source=supriya.ugens.WhiteNoise.ar(),
        ... )
        >>> pv_mag_freeze = supriya.ugens.PV_MagFreeze.kr(
        ...     pv_chain=pv_chain,
        ...     freeze=0,
        ... )
        >>> pv_mag_freeze
        <PV_MagFreeze.kr()[0]>
    """

    pv_chain = param()
    freeze = param(0)


@ugen(kr=True, is_width_first=True)
class PV_MagMul(PV_ChainUGen):
    """
    Multiplies FFT magnitudes.

    ::

        >>> pv_chain_a = supriya.ugens.FFT.kr(
        ...     source=supriya.ugens.WhiteNoise.ar(),
        ... )
        >>> pv_chain_b = supriya.ugens.FFT.kr(
        ...     source=supriya.ugens.LFSaw.ar(),
        ... )
        >>> pv_mag_mul = supriya.ugens.PV_MagMul.kr(
        ...     pv_chain_a=pv_chain_a,
        ...     pv_chain_b=pv_chain_b,
        ... )
        >>> pv_mag_mul
        <PV_MagMul.kr()[0]>
    """

    pv_chain_a = param()
    pv_chain_b = param()


@ugen(kr=True, is_width_first=True)
class PV_MagNoise(PV_ChainUGen):
    """
    Multiplies magnitudes by noise.

    ::

        >>> pv_chain = supriya.ugens.FFT.kr(
        ...     source=supriya.ugens.WhiteNoise.ar(),
        ... )
        >>> pv_mag_noise = supriya.ugens.PV_MagNoise.kr(
        ...     pv_chain=pv_chain,
        ... )
        >>> pv_mag_noise
        <PV_MagNoise.kr()[0]>
    """

    pv_chain = param()


@ugen(kr=True, is_width_first=True)
class PV_MagShift(PV_ChainUGen):
    """
    Shifts and stretches magnitude bin position.

    ::

        >>> pv_chain = supriya.ugens.FFT.kr(
        ...     source=supriya.ugens.WhiteNoise.ar(),
        ... )
        >>> pv_mag_shift = supriya.ugens.PV_MagShift.kr(
        ...     pv_chain=pv_chain,
        ...     shift=0,
        ...     stretch=1,
        ... )
        >>> pv_mag_shift
        <PV_MagShift.kr()[0]>
    """

    pv_chain = param()
    stretch = param(1.0)
    shift = param(0.0)


@ugen(kr=True, is_width_first=True)
class PV_MagSmear(PV_ChainUGen):
    """
    Averages magnitudes across bins.

    ::

        >>> pv_chain = supriya.ugens.FFT.kr(
        ...     source=supriya.ugens.WhiteNoise.ar(),
        ... )
        >>> pv_mag_smear = supriya.ugens.PV_MagSmear.kr(
        ...     bins=0,
        ...     pv_chain=pv_chain,
        ... )
        >>> pv_mag_smear
        <PV_MagSmear.kr()[0]>
    """

    pv_chain = param()
    bins = param(0)


@ugen(kr=True, is_width_first=True)
class PV_MagSquared(PV_ChainUGen):
    """
    Squares magnitudes.

    ::

        >>> pv_chain = supriya.ugens.FFT.kr(
        ...     source=supriya.ugens.WhiteNoise.ar(),
        ... )
        >>> pv_mag_squared = supriya.ugens.PV_MagSquared.kr(
        ...     pv_chain=pv_chain,
        ... )
        >>> pv_mag_squared
        <PV_MagSquared.kr()[0]>
    """

    pv_chain = param()


@ugen(kr=True, is_width_first=True)
class PV_Max(PV_ChainUGen):
    """
    Maximum magnitude.

    ::

        >>> pv_chain_a = supriya.ugens.FFT.kr(
        ...     source=supriya.ugens.WhiteNoise.ar(),
        ... )
        >>> pv_chain_b = supriya.ugens.FFT.kr(
        ...     source=supriya.ugens.LFSaw.ar(),
        ... )
        >>> pv_max = supriya.ugens.PV_Max.kr(
        ...     pv_chain_a=pv_chain_a,
        ...     pv_chain_b=pv_chain_b,
        ... )
        >>> pv_max
        <PV_Max.kr()[0]>
    """

    pv_chain_a = param()
    pv_chain_b = param()


@ugen(kr=True, is_width_first=True)
class PV_Min(PV_ChainUGen):
    """
    Minimum magnitude.

    ::

        >>> pv_chain_a = supriya.ugens.FFT.kr(
        ...     source=supriya.ugens.WhiteNoise.ar(),
        ... )
        >>> pv_chain_b = supriya.ugens.FFT.kr(
        ...     source=supriya.ugens.LFSaw.ar(),
        ... )
        >>> pv_min = supriya.ugens.PV_Min.kr(
        ...     pv_chain_a=pv_chain_a,
        ...     pv_chain_b=pv_chain_b,
        ... )
        >>> pv_min
        <PV_Min.kr()[0]>
    """

    pv_chain_a = param()
    pv_chain_b = param()


@ugen(kr=True, is_width_first=True)
class PV_Mul(PV_ChainUGen):
    """
    Complex multiplication.

    ::

        >>> pv_chain_a = supriya.ugens.FFT.kr(
        ...     source=supriya.ugens.WhiteNoise.ar(),
        ... )
        >>> pv_chain_b = supriya.ugens.FFT.kr(
        ...     source=supriya.ugens.LFSaw.ar(),
        ... )
        >>> pv_mul = supriya.ugens.PV_Mul.kr(
        ...     pv_chain_a=pv_chain_a,
        ...     pv_chain_b=pv_chain_b,
        ... )
        >>> pv_mul
        <PV_Mul.kr()[0]>
    """

    pv_chain_a = param()
    pv_chain_b = param()


@ugen(kr=True, is_width_first=True)
class PV_PhaseShift(PV_ChainUGen):
    """
    Shifts phase.

    ::

        >>> pv_chain = supriya.ugens.FFT.kr(
        ...     source=supriya.ugens.WhiteNoise.ar(),
        ... )
        >>> shift = supriya.ugens.LFNoise2.kr(frequency=1).scale(-1, 1, -180, 180)
        >>> pv_phase_shift = supriya.ugens.PV_PhaseShift.kr(
        ...     pv_chain=pv_chain,
        ...     integrate=0,
        ...     shift=shift,
        ... )
        >>> pv_phase_shift
        <PV_PhaseShift.kr()[0]>
    """

    pv_chain = param()
    shift = param()
    integrate = param(0)


@ugen(kr=True, is_width_first=True)
class PV_PhaseShift270(PV_ChainUGen):
    """
    Shifts phase by 270 degrees.

    ::

        >>> pv_chain = supriya.ugens.FFT.kr(
        ...     source=supriya.ugens.WhiteNoise.ar(),
        ... )
        >>> pv_phase_shift_270 = supriya.ugens.PV_PhaseShift270.kr(
        ...     pv_chain=pv_chain,
        ... )
        >>> pv_phase_shift_270
        <PV_PhaseShift270.kr()[0]>
    """

    pv_chain = param()


@ugen(kr=True, is_width_first=True)
class PV_PhaseShift90(PV_ChainUGen):
    """
    Shifts phase by 90 degrees.

    ::

        >>> pv_chain = supriya.ugens.FFT.kr(
        ...     source=supriya.ugens.WhiteNoise.ar(),
        ... )
        >>> pv_phase_shift_90 = supriya.ugens.PV_PhaseShift90.kr(
        ...     pv_chain=pv_chain,
        ... )
        >>> pv_phase_shift_90
        <PV_PhaseShift90.kr()[0]>
    """

    pv_chain = param()


@ugen(kr=True, is_width_first=True)
class PV_RandComb(PV_ChainUGen):
    """
    Passes random bins.

    ::

        >>> pv_chain = supriya.ugens.FFT.kr(
        ...     source=supriya.ugens.WhiteNoise.ar(),
        ... )
        >>> pv_rand_comb = supriya.ugens.PV_RandComb.kr(
        ...     pv_chain=pv_chain,
        ...     trigger=0,
        ...     wipe=0,
        ... )
        >>> pv_rand_comb
        <PV_RandComb.kr()[0]>
    """

    pv_chain = param()
    wipe = param(0)
    trigger = param(0)


@ugen(kr=True, is_width_first=True)
class PV_RandWipe(PV_ChainUGen):
    """
    Crossfades in random bin order.

    ::

        >>> pv_chain_a = supriya.ugens.FFT.kr(
        ...     source=supriya.ugens.WhiteNoise.ar(),
        ... )
        >>> pv_chain_b = supriya.ugens.FFT.kr(
        ...     source=supriya.ugens.LFSaw.ar(),
        ... )
        >>> pv_rand_wipe = supriya.ugens.PV_RandWipe.kr(
        ...     pv_chain_a=pv_chain_a,
        ...     pv_chain_b=pv_chain_b,
        ...     trigger=0,
        ...     wipe=0,
        ... )
        >>> pv_rand_wipe
        <PV_RandWipe.kr()[0]>
    """

    pv_chain_a = param()
    pv_chain_b = param()
    wipe = param(0)
    trigger = param(0)


@ugen(kr=True, is_width_first=True)
class PV_RectComb(PV_ChainUGen):
    """
    Makes gaps in the spectrum.

    ::

        >>> pv_chain = supriya.ugens.FFT.kr(
        ...     source=supriya.ugens.WhiteNoise.ar(),
        ... )
        >>> pv_rect_comb = supriya.ugens.PV_RectComb.kr(
        ...     pv_chain=pv_chain,
        ...     num_teeth=0,
        ...     phase=0,
        ...     width=0.5,
        ... )
        >>> pv_rect_comb
        <PV_RectComb.kr()[0]>
    """

    pv_chain = param()
    num_teeth = param(0)
    phase = param(0)
    width = param(0.5)


@ugen(kr=True, is_width_first=True)
class PV_RectComb2(PV_ChainUGen):
    """
    Makes gaps in the spectrum.

    ::

        >>> pv_chain_a = supriya.ugens.FFT.kr(
        ...     source=supriya.ugens.WhiteNoise.ar(),
        ... )
        >>> pv_chain_b = supriya.ugens.FFT.kr(
        ...     source=supriya.ugens.LFSaw.ar(),
        ... )
        >>> pv_rect_comb_2 = supriya.ugens.PV_RectComb2.kr(
        ...     pv_chain_a=pv_chain_a,
        ...     pv_chain_b=pv_chain_b,
        ...     num_teeth=0,
        ...     phase=0,
        ...     width=0.5,
        ... )
        >>> pv_rect_comb_2
        <PV_RectComb2.kr()[0]>
    """

    pv_chain_a = param()
    pv_chain_b = param()
    num_teeth = param(0)
    phase = param(0)
    width = param(0.5)


@ugen(ar=True, kr=True)
class RunningSum(UGen):
    """
    Tracks running sum over ``n`` frames.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> running_sum = supriya.ugens.RunningSum.ar(
        ...     sample_count=40,
        ...     source=source,
        ... )
        >>> running_sum
        <RunningSum.ar()[0]>
    """

    source = param()
    sample_count = param(40)



from typing import Any, Dict, Tuple

from ..enums import CalculationRate
from .core import UGen, param, ugen


@ugen(ar=True, kr=True)
class CheckBadValues(UGen):
    """
    Tests for infinity, not-a-number, and denormals.

    ::

        >>> source = supriya.ugens.SinOsc.ar()
        >>> ugen_id = 23
        >>> post_mode = 0
        >>> check_bad_values = supriya.ugens.CheckBadValues.ar(
        ...     source=source,
        ...     ugen_id=ugen_id,
        ...     post_mode=post_mode,
        ... )
        >>> check_bad_values
        <CheckBadValues.ar()[0]>
    """

    source = param()
    ugen_id = param(0)
    post_mode = param(2)

    def _postprocess_kwargs(
        self,
        *,
        calculation_rate: CalculationRate,
        **kwargs,
    ) -> Tuple[CalculationRate, Dict[str, Any]]:
        if kwargs["post_mode"] not in (0, 1, 2):
            raise ValueError(f"Invalid post mode: {kwargs['post_mode']}")
        return calculation_rate, kwargs


@ugen(ar=True, kr=True)
class Sanitize(UGen):
    """
    Remove infinity, NaN, and denormals.
    """

    source = param()
    replace = param(0.0)



from ..enums import CalculationRate
from .core import PseudoUGen, UGen, param, ugen
from .delay import DelayN


@ugen(ar=True, kr=True)
class Amplitude(UGen):
    """
    An amplitude follower.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> amplitude = supriya.ugens.Amplitude.kr(
        ...     attack_time=0.01,
        ...     release_time=0.01,
        ...     source=source,
        ... )
        >>> amplitude
        <Amplitude.kr()[0]>
    """

    source = param()
    attack_time = param(0.01)
    release_time = param(0.01)


@ugen(ar=True)
class Compander(UGen):
    """
    A general purpose hard-knee dynamics processor.
    """

    source = param()
    control = param(0.0)
    threshold = param(0.5)
    slope_below = param(1.0)
    slope_above = param(1.0)
    clamp_time = param(0.01)
    relax_time = param(0.1)


class CompanderD(PseudoUGen):
    """
    A convenience constructor for Compander.
    """

    ### PUBLIC METHODS ###

    @classmethod
    def ar(
        cls,
        source=None,
        threshold=0.5,
        clamp_time=0.01,
        relax_time=0.1,
        slope_above=1.0,
        slope_below=1.0,
    ):
        """
        Constructs an audio-rate dynamics processor.

        .. container:: example

            ::

                >>> source = supriya.ugens.In.ar(bus=0)
                >>> compander_d = supriya.ugens.CompanderD.ar(
                ...     source=source,
                ... )
                >>> supriya.graph(compander_d)  # doctest: +SKIP

            ::

                >>> print(compander_d)
                synthdef:
                    name: ...
                    ugens:
                    -   In.ar:
                            bus: 0.0
                    -   DelayN.ar:
                            source: In.ar[0]
                            maximum_delay_time: 0.01
                            delay_time: 0.01
                    -   Compander.ar:
                            source: In.ar[0]
                            control: DelayN.ar[0]
                            threshold: 0.5
                            slope_below: 1.0
                            slope_above: 1.0
                            clamp_time: 0.01
                            relax_time: 0.1

        Returns ugen graph.
        """
        control = DelayN.ar(
            source=source, maximum_delay_time=clamp_time, delay_time=clamp_time
        )
        return Compander._new_expanded(
            clamp_time=clamp_time,
            calculation_rate=CalculationRate.AUDIO,
            relax_time=relax_time,
            slope_above=slope_above,
            slope_below=slope_below,
            source=source,
            control=control,
            threshold=threshold,
        )


@ugen(ar=True)
class Limiter(UGen):
    """
    A peak limiter.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> limiter = supriya.ugens.Limiter.ar(
        ...     duration=0.01,
        ...     level=1,
        ...     source=source,
        ... )
        >>> limiter
        <Limiter.ar()[0]>
    """

    source = param()
    level = param(1.0)
    duration = param(0.01)


@ugen(ar=True)
class Normalizer(UGen):
    """
    A dynamics flattener.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> normalizer = supriya.ugens.Normalizer.ar(
        ...     duration=0.01,
        ...     level=1,
        ...     source=source,
        ... )
        >>> normalizer
        <Normalizer.ar()[0]>
    """

    source = param()
    level = param(1.0)
    duration = param(0.01)



import asyncio
import logging
import platform
import random

import pytest

from supriya.contexts.realtime import (
    DEFAULT_HEALTHCHECK,
    AsyncServer,
    Server,
    ServerLifecycleEvent,
)
from supriya.enums import BootStatus
from supriya.exceptions import (
    OwnedServerShutdown,
    ServerCannotBoot,
    ServerOnline,
    TooManyClients,
    UnownedServerShutdown,
)
from supriya.osc import find_free_port
from supriya.scsynth import kill

supernova = pytest.param(
    "supernova",
    marks=pytest.mark.skipif(
        platform.system() == "Windows", reason="Supernova won't boot on Windows"
    ),
)


def setup_context(context_class, *, async_callback=False, name=None):
    def on_event(event: ServerLifecycleEvent) -> None:
        events.append(event)

    async def on_event_async(event: ServerLifecycleEvent) -> None:
        await asyncio.sleep(0)
        events.append(event)

    events = []
    context = context_class(name=name)
    callback = on_event_async if async_callback else on_event
    for event in ServerLifecycleEvent:
        context.on(event, callback)
    return context, events


async def get(x):
    if asyncio.iscoroutine(x):
        return await x
    return x


async def get_future(x, timeout=10):
    if asyncio.isfuture(x):
        return await asyncio.wait_for(x, timeout)
    return x.result(timeout)


logger = logging.getLogger(__name__)


@pytest.fixture(autouse=True)
def use_caplog(caplog):
    caplog.set_level(logging.INFO)


@pytest.fixture(autouse=True)
def healthcheck_attempts(monkeypatch):
    monkeypatch.setattr(DEFAULT_HEALTHCHECK, "max_attempts", 1)


@pytest.mark.asyncio
@pytest.mark.parametrize("executable", ["scsynth", supernova])
@pytest.mark.parametrize(
    "async_callback, context_class",
    [(False, AsyncServer), (True, AsyncServer), (False, Server)],
)
async def test_boot_only(async_callback, context_class, executable):
    context, events = setup_context(context_class, async_callback=async_callback)
    assert context.boot_status == BootStatus.OFFLINE
    assert not context.is_owner
    #
    await get(context.boot(executable=executable))
    assert context.boot_status == BootStatus.ONLINE
    assert context.is_owner
    assert events == [
        ServerLifecycleEvent.BOOTING,
        ServerLifecycleEvent.PROCESS_BOOTED,
        ServerLifecycleEvent.CONNECTING,
        ServerLifecycleEvent.OSC_CONNECTED,
        ServerLifecycleEvent.CONNECTED,
        ServerLifecycleEvent.BOOTED,
    ]
    assert context.boot_future.done()
    assert not context.exit_future.done()


@pytest.mark.asyncio
@pytest.mark.parametrize("executable", ["scsynth", supernova])
@pytest.mark.parametrize(
    "async_callback, context_class",
    [(False, AsyncServer), (True, AsyncServer), (False, Server)],
)
async def test_boot_and_quit(async_callback, context_class, executable):
    context, events = setup_context(context_class, async_callback=async_callback)
    assert context.boot_status == BootStatus.OFFLINE
    assert not context.is_owner
    #
    await get(context.boot(executable=executable))
    assert context.boot_status == BootStatus.ONLINE
    assert context.is_owner
    assert events == [
        ServerLifecycleEvent.BOOTING,
        ServerLifecycleEvent.PROCESS_BOOTED,
        ServerLifecycleEvent.CONNECTING,
        ServerLifecycleEvent.OSC_CONNECTED,
        ServerLifecycleEvent.CONNECTED,
        ServerLifecycleEvent.BOOTED,
    ]
    #
    await get(context.quit())
    assert context.boot_status == BootStatus.OFFLINE
    assert not context.is_owner
    assert events == [
        ServerLifecycleEvent.BOOTING,
        ServerLifecycleEvent.PROCESS_BOOTED,
        ServerLifecycleEvent.CONNECTING,
        ServerLifecycleEvent.OSC_CONNECTED,
        ServerLifecycleEvent.CONNECTED,
        ServerLifecycleEvent.BOOTED,
        ServerLifecycleEvent.QUITTING,
        ServerLifecycleEvent.DISCONNECTING,
        ServerLifecycleEvent.OSC_DISCONNECTED,
        ServerLifecycleEvent.DISCONNECTED,
        ServerLifecycleEvent.PROCESS_QUIT,
        ServerLifecycleEvent.QUIT,
    ]
    assert context.boot_future.done()
    assert context.exit_future.done()


@pytest.mark.asyncio
@pytest.mark.parametrize("executable", ["scsynth", supernova])
@pytest.mark.parametrize(
    "async_callback, context_class",
    [(False, AsyncServer), (True, AsyncServer), (False, Server)],
)
async def test_boot_and_boot(async_callback, context_class, executable):
    context, events = setup_context(context_class, async_callback=async_callback)
    assert context.boot_status == BootStatus.OFFLINE
    assert not context.is_owner
    #
    await get(context.boot(executable=executable))
    assert context.boot_status == BootStatus.ONLINE
    assert context.is_owner
    assert events == [
        ServerLifecycleEvent.BOOTING,
        ServerLifecycleEvent.PROCESS_BOOTED,
        ServerLifecycleEvent.CONNECTING,
        ServerLifecycleEvent.OSC_CONNECTED,
        ServerLifecycleEvent.CONNECTED,
        ServerLifecycleEvent.BOOTED,
    ]
    #
    with pytest.raises(ServerOnline):
        await get(context.boot(executable=executable))
    assert context.boot_status == BootStatus.ONLINE
    assert context.is_owner
    assert events == [
        ServerLifecycleEvent.BOOTING,
        ServerLifecycleEvent.PROCESS_BOOTED,
        ServerLifecycleEvent.CONNECTING,
        ServerLifecycleEvent.OSC_CONNECTED,
        ServerLifecycleEvent.CONNECTED,
        ServerLifecycleEvent.BOOTED,
    ]


@pytest.mark.asyncio
@pytest.mark.parametrize("executable", ["scsynth", supernova])
@pytest.mark.parametrize(
    "async_callback, context_class",
    [(False, AsyncServer), (True, AsyncServer), (False, Server)],
)
async def test_boot_and_quit_and_quit(async_callback, context_class, executable):
    context, events = setup_context(context_class, async_callback=async_callback)
    assert context.boot_status == BootStatus.OFFLINE
    assert not context.is_owner
    #
    await get(context.boot(executable=executable))
    assert context.boot_status == BootStatus.ONLINE
    assert context.is_owner
    assert events == [
        ServerLifecycleEvent.BOOTING,
        ServerLifecycleEvent.PROCESS_BOOTED,
        ServerLifecycleEvent.CONNECTING,
        ServerLifecycleEvent.OSC_CONNECTED,
        ServerLifecycleEvent.CONNECTED,
        ServerLifecycleEvent.BOOTED,
    ]
    #
    await get(context.quit())
    assert context.boot_status == BootStatus.OFFLINE
    assert not context.is_owner
    assert events == [
        ServerLifecycleEvent.BOOTING,
        ServerLifecycleEvent.PROCESS_BOOTED,
        ServerLifecycleEvent.CONNECTING,
        ServerLifecycleEvent.OSC_CONNECTED,
        ServerLifecycleEvent.CONNECTED,
        ServerLifecycleEvent.BOOTED,
        ServerLifecycleEvent.QUITTING,
        ServerLifecycleEvent.DISCONNECTING,
        ServerLifecycleEvent.OSC_DISCONNECTED,
        ServerLifecycleEvent.DISCONNECTED,
        ServerLifecycleEvent.PROCESS_QUIT,
        ServerLifecycleEvent.QUIT,
    ]
    #
    await get(context.quit())
    assert context.boot_status == BootStatus.OFFLINE
    assert not context.is_owner
    assert events == [
        ServerLifecycleEvent.BOOTING,
        ServerLifecycleEvent.PROCESS_BOOTED,
        ServerLifecycleEvent.CONNECTING,
        ServerLifecycleEvent.OSC_CONNECTED,
        ServerLifecycleEvent.CONNECTED,
        ServerLifecycleEvent.BOOTED,
        ServerLifecycleEvent.QUITTING,
        ServerLifecycleEvent.DISCONNECTING,
        ServerLifecycleEvent.OSC_DISCONNECTED,
        ServerLifecycleEvent.DISCONNECTED,
        ServerLifecycleEvent.PROCESS_QUIT,
        ServerLifecycleEvent.QUIT,
    ]


@pytest.mark.asyncio
@pytest.mark.parametrize("executable", ["scsynth", supernova])
@pytest.mark.parametrize(
    "async_callback, context_class",
    [(False, AsyncServer), (True, AsyncServer), (False, Server)],
)
async def test_boot_and_connect(async_callback, context_class, executable):
    context, events = setup_context(context_class, async_callback=async_callback)
    assert context.boot_status == BootStatus.OFFLINE
    assert not context.is_owner
    #
    await get(context.boot(executable=executable))
    assert context.boot_status == BootStatus.ONLINE
    assert context.is_owner
    assert events == [
        ServerLifecycleEvent.BOOTING,
        ServerLifecycleEvent.PROCESS_BOOTED,
        ServerLifecycleEvent.CONNECTING,
        ServerLifecycleEvent.OSC_CONNECTED,
        ServerLifecycleEvent.CONNECTED,
        ServerLifecycleEvent.BOOTED,
    ]
    #
    with pytest.raises(ServerOnline):
        await get(context.connect())
    assert context.boot_status == BootStatus.ONLINE
    assert context.is_owner
    assert events == [
        ServerLifecycleEvent.BOOTING,
        ServerLifecycleEvent.PROCESS_BOOTED,
        ServerLifecycleEvent.CONNECTING,
        ServerLifecycleEvent.OSC_CONNECTED,
        ServerLifecycleEvent.CONNECTED,
        ServerLifecycleEvent.BOOTED,
    ]


@pytest.mark.asyncio
@pytest.mark.parametrize("executable", ["scsynth", supernova])
@pytest.mark.parametrize(
    "async_callback, context_class",
    [(False, AsyncServer), (True, AsyncServer), (False, Server)],
)
async def test_boot_a_and_boot_b_cannot_boot(async_callback, context_class, executable):
    context_a, events_a = setup_context(
        context_class, async_callback=async_callback, name="one"
    )
    context_b, events_b = setup_context(
        context_class, async_callback=async_callback, name="two"
    )
    assert context_a.boot_status == BootStatus.OFFLINE and not context_a.is_owner
    assert context_b.boot_status == BootStatus.OFFLINE and not context_b.is_owner
    #
    await get(context_a.boot(maximum_logins=4, executable=executable))
    assert context_a.boot_status == BootStatus.ONLINE and context_a.is_owner
    assert context_b.boot_status == BootStatus.OFFLINE and not context_b.is_owner
    assert events_a == [
        ServerLifecycleEvent.BOOTING,
        ServerLifecycleEvent.PROCESS_BOOTED,
        ServerLifecycleEvent.CONNECTING,
        ServerLifecycleEvent.OSC_CONNECTED,
        ServerLifecycleEvent.CONNECTED,
        ServerLifecycleEvent.BOOTED,
    ]
    assert events_b == []
    #
    with pytest.raises(ServerCannotBoot):
        await get(context_b.boot(maximum_logins=4, executable=executable))
    assert context_a.boot_status == BootStatus.ONLINE and context_a.is_owner
    assert context_b.boot_status == BootStatus.OFFLINE and not context_b.is_owner
    assert events_a == [
        ServerLifecycleEvent.BOOTING,
        ServerLifecycleEvent.PROCESS_BOOTED,
        ServerLifecycleEvent.CONNECTING,
        ServerLifecycleEvent.OSC_CONNECTED,
        ServerLifecycleEvent.CONNECTED,
        ServerLifecycleEvent.BOOTED,
    ]
    assert events_b == [
        ServerLifecycleEvent.BOOTING,
        ServerLifecycleEvent.PROCESS_PANICKED,
    ]


# scsynth only
@pytest.mark.asyncio
@pytest.mark.parametrize("executable", ["scsynth", supernova])
@pytest.mark.parametrize(
    "async_callback, context_class",
    [(False, AsyncServer), (True, AsyncServer), (False, Server)],
)
async def test_boot_a_and_connect_b_too_many_clients(
    async_callback, context_class, executable
):
    context_a, events_a = setup_context(
        context_class, async_callback=async_callback, name="one"
    )
    context_b, events_b = setup_context(
        context_class, async_callback=async_callback, name="two"
    )
    assert context_a.boot_status == BootStatus.OFFLINE and not context_a.is_owner
    assert context_b.boot_status == BootStatus.OFFLINE and not context_b.is_owner
    #
    await get(context_a.boot(maximum_logins=1))
    assert context_a.boot_status == BootStatus.ONLINE and context_a.is_owner
    assert context_b.boot_status == BootStatus.OFFLINE and not context_b.is_owner
    assert events_a == [
        ServerLifecycleEvent.BOOTING,
        ServerLifecycleEvent.PROCESS_BOOTED,
        ServerLifecycleEvent.CONNECTING,
        ServerLifecycleEvent.OSC_CONNECTED,
        ServerLifecycleEvent.CONNECTED,
        ServerLifecycleEvent.BOOTED,
    ]
    assert events_b == []
    #
    with pytest.raises(TooManyClients):
        await get(context_b.connect())
    assert context_a.boot_status == BootStatus.ONLINE and context_a.is_owner
    assert context_b.boot_status == BootStatus.OFFLINE and not context_b.is_owner
    assert events_a == [
        ServerLifecycleEvent.BOOTING,
        ServerLifecycleEvent.PROCESS_BOOTED,
        ServerLifecycleEvent.CONNECTING,
        ServerLifecycleEvent.OSC_CONNECTED,
        ServerLifecycleEvent.CONNECTED,
        ServerLifecycleEvent.BOOTED,
    ]
    assert events_b == [
        ServerLifecycleEvent.CONNECTING,
        ServerLifecycleEvent.DISCONNECTING,
        ServerLifecycleEvent.DISCONNECTED,
    ]


@pytest.mark.asyncio
@pytest.mark.parametrize("executable", ["scsynth", supernova])
@pytest.mark.parametrize(
    "async_callback, context_class",
    [(False, AsyncServer), (True, AsyncServer), (False, Server)],
)
async def test_boot_a_and_connect_b_and_quit_a(
    async_callback, context_class, executable
):
    logger.warning("START")
    context_a, events_a = setup_context(
        context_class, async_callback=async_callback, name="one"
    )
    context_b, events_b = setup_context(
        context_class, async_callback=async_callback, name="two"
    )
    assert context_a.boot_status == BootStatus.OFFLINE and not context_a.is_owner
    assert context_b.boot_status == BootStatus.OFFLINE and not context_b.is_owner
    #
    logger.warning("BOOT A")
    await get(context_a.boot(maximum_logins=2, executable=executable))
    assert events_a == [
        ServerLifecycleEvent.BOOTING,
        ServerLifecycleEvent.PROCESS_BOOTED,
        ServerLifecycleEvent.CONNECTING,
        ServerLifecycleEvent.OSC_CONNECTED,
        ServerLifecycleEvent.CONNECTED,
        ServerLifecycleEvent.BOOTED,
    ]
    assert events_b == []
    #
    logger.warning("CONNECT B")
    await get(context_b.connect())
    assert context_a.boot_status == BootStatus.ONLINE and context_a.is_owner
    assert context_b.boot_status == BootStatus.ONLINE and not context_b.is_owner
    assert events_a == [
        ServerLifecycleEvent.BOOTING,
        ServerLifecycleEvent.PROCESS_BOOTED,
        ServerLifecycleEvent.CONNECTING,
        ServerLifecycleEvent.OSC_CONNECTED,
        ServerLifecycleEvent.CONNECTED,
        ServerLifecycleEvent.BOOTED,
    ]
    assert events_b == [
        ServerLifecycleEvent.CONNECTING,
        ServerLifecycleEvent.OSC_CONNECTED,
        ServerLifecycleEvent.CONNECTED,
    ]
    #
    logger.warning("PROCESS_QUIT A")
    await get(context_a.quit())
    assert context_a.boot_status == BootStatus.OFFLINE and not context_a.is_owner
    assert events_a == [
        ServerLifecycleEvent.BOOTING,
        ServerLifecycleEvent.PROCESS_BOOTED,
        ServerLifecycleEvent.CONNECTING,
        ServerLifecycleEvent.OSC_CONNECTED,
        ServerLifecycleEvent.CONNECTED,
        ServerLifecycleEvent.BOOTED,
        ServerLifecycleEvent.QUITTING,
        ServerLifecycleEvent.DISCONNECTING,
        ServerLifecycleEvent.OSC_DISCONNECTED,
        ServerLifecycleEvent.DISCONNECTED,
        ServerLifecycleEvent.PROCESS_QUIT,
        ServerLifecycleEvent.QUIT,
    ]
    assert events_b == [
        ServerLifecycleEvent.CONNECTING,
        ServerLifecycleEvent.OSC_CONNECTED,
        ServerLifecycleEvent.CONNECTED,
    ]
    #
    logger.warning("AWAIT B")
    await get_future(context_b.exit_future)
    logger.warning("DONE AWAITING B")
    assert context_b.boot_status == BootStatus.OFFLINE and not context_b.is_owner
    assert events_a == [
        ServerLifecycleEvent.BOOTING,
        ServerLifecycleEvent.PROCESS_BOOTED,
        ServerLifecycleEvent.CONNECTING,
        ServerLifecycleEvent.OSC_CONNECTED,
        ServerLifecycleEvent.CONNECTED,
        ServerLifecycleEvent.BOOTED,
        ServerLifecycleEvent.QUITTING,
        ServerLifecycleEvent.DISCONNECTING,
        ServerLifecycleEvent.OSC_DISCONNECTED,
        ServerLifecycleEvent.DISCONNECTED,
        ServerLifecycleEvent.PROCESS_QUIT,
        ServerLifecycleEvent.QUIT,
    ]
    assert events_b == [
        ServerLifecycleEvent.CONNECTING,
        ServerLifecycleEvent.OSC_CONNECTED,
        ServerLifecycleEvent.CONNECTED,
        ServerLifecycleEvent.OSC_PANICKED,
        ServerLifecycleEvent.DISCONNECTING,
        ServerLifecycleEvent.DISCONNECTED,
    ]
    logger.warning("END")


@pytest.mark.asyncio
@pytest.mark.parametrize("executable", ["scsynth", supernova])
@pytest.mark.parametrize(
    "async_callback, context_class",
    [(False, AsyncServer), (True, AsyncServer), (False, Server)],
)
async def test_boot_a_and_connect_b_and_disconnect_b(
    async_callback, context_class, executable
):
    context_a, events_a = setup_context(
        context_class, async_callback=async_callback, name="one"
    )
    context_b, events_b = setup_context(
        context_class, async_callback=async_callback, name="two"
    )
    assert context_a.boot_status == BootStatus.OFFLINE and not context_a.is_owner
    assert context_b.boot_status == BootStatus.OFFLINE and not context_b.is_owner
    #
    await get(context_a.boot(maximum_logins=2, executable=executable))
    await get(context_b.connect())
    assert context_a.boot_status == BootStatus.ONLINE and context_a.is_owner
    assert context_b.boot_status == BootStatus.ONLINE and not context_b.is_owner
    assert events_a == [
        ServerLifecycleEvent.BOOTING,
        ServerLifecycleEvent.PROCESS_BOOTED,
        ServerLifecycleEvent.CONNECTING,
        ServerLifecycleEvent.OSC_CONNECTED,
        ServerLifecycleEvent.CONNECTED,
        ServerLifecycleEvent.BOOTED,
    ]
    assert events_b == [
        ServerLifecycleEvent.CONNECTING,
        ServerLifecycleEvent.OSC_CONNECTED,
        ServerLifecycleEvent.CONNECTED,
    ]
    #
    await get(context_b.disconnect())
    assert context_a.boot_status == BootStatus.ONLINE and context_a.is_owner
    assert context_b.boot_status == BootStatus.OFFLINE and not context_b.is_owner
    assert events_a == [
        ServerLifecycleEvent.BOOTING,
        ServerLifecycleEvent.PROCESS_BOOTED,
        ServerLifecycleEvent.CONNECTING,
        ServerLifecycleEvent.OSC_CONNECTED,
        ServerLifecycleEvent.CONNECTED,
        ServerLifecycleEvent.BOOTED,
    ]
    assert events_b == [
        ServerLifecycleEvent.CONNECTING,
        ServerLifecycleEvent.OSC_CONNECTED,
        ServerLifecycleEvent.CONNECTED,
        ServerLifecycleEvent.DISCONNECTING,
        ServerLifecycleEvent.OSC_DISCONNECTED,
        ServerLifecycleEvent.DISCONNECTED,
    ]


@pytest.mark.asyncio
@pytest.mark.parametrize("executable", ["scsynth", supernova])
@pytest.mark.parametrize(
    "async_callback, context_class",
    [(False, AsyncServer), (True, AsyncServer), (False, Server)],
)
async def test_boot_a_and_connect_b_and_disconnect_a(
    async_callback, context_class, executable
):
    context_a, events_a = setup_context(
        context_class, async_callback=async_callback, name="one"
    )
    context_b, events_b = setup_context(
        context_class, async_callback=async_callback, name="two"
    )
    assert context_a.boot_status == BootStatus.OFFLINE and not context_a.is_owner
    assert context_b.boot_status == BootStatus.OFFLINE and not context_b.is_owner
    #
    await get(context_a.boot(maximum_logins=2, executable=executable))
    await get(context_b.connect())
    assert context_a.boot_status == BootStatus.ONLINE and context_a.is_owner
    assert context_b.boot_status == BootStatus.ONLINE and not context_b.is_owner
    assert events_a == [
        ServerLifecycleEvent.BOOTING,
        ServerLifecycleEvent.PROCESS_BOOTED,
        ServerLifecycleEvent.CONNECTING,
        ServerLifecycleEvent.OSC_CONNECTED,
        ServerLifecycleEvent.CONNECTED,
        ServerLifecycleEvent.BOOTED,
    ]
    assert events_b == [
        ServerLifecycleEvent.CONNECTING,
        ServerLifecycleEvent.OSC_CONNECTED,
        ServerLifecycleEvent.CONNECTED,
    ]
    #
    with pytest.raises(OwnedServerShutdown):
        await get(context_a.disconnect())
    assert context_a.boot_status == BootStatus.ONLINE and context_a.is_owner
    assert context_b.boot_status == BootStatus.ONLINE and not context_b.is_owner
    assert events_a == [
        ServerLifecycleEvent.BOOTING,
        ServerLifecycleEvent.PROCESS_BOOTED,
        ServerLifecycleEvent.CONNECTING,
        ServerLifecycleEvent.OSC_CONNECTED,
        ServerLifecycleEvent.CONNECTED,
        ServerLifecycleEvent.BOOTED,
    ]
    assert events_b == [
        ServerLifecycleEvent.CONNECTING,
        ServerLifecycleEvent.OSC_CONNECTED,
        ServerLifecycleEvent.CONNECTED,
    ]


@pytest.mark.asyncio
@pytest.mark.parametrize("executable", ["scsynth", supernova])
@pytest.mark.parametrize(
    "async_callback, context_class",
    [(False, AsyncServer), (True, AsyncServer), (False, Server)],
)
async def test_boot_a_and_connect_b_and_quit_b(
    async_callback, context_class, executable
):
    context_a, events_a = setup_context(
        context_class, async_callback=async_callback, name="one"
    )
    context_b, events_b = setup_context(
        context_class, async_callback=async_callback, name="two"
    )
    assert context_a.boot_status == BootStatus.OFFLINE and not context_a.is_owner
    assert context_b.boot_status == BootStatus.OFFLINE and not context_b.is_owner
    #
    await get(context_a.boot(maximum_logins=2, executable=executable))
    await get(context_b.connect())
    assert context_a.boot_status == BootStatus.ONLINE and context_a.is_owner
    assert context_b.boot_status == BootStatus.ONLINE and not context_b.is_owner
    assert events_a == [
        ServerLifecycleEvent.BOOTING,
        ServerLifecycleEvent.PROCESS_BOOTED,
        ServerLifecycleEvent.CONNECTING,
        ServerLifecycleEvent.OSC_CONNECTED,
        ServerLifecycleEvent.CONNECTED,
        ServerLifecycleEvent.BOOTED,
    ]
    assert events_b == [
        ServerLifecycleEvent.CONNECTING,
        ServerLifecycleEvent.OSC_CONNECTED,
        ServerLifecycleEvent.CONNECTED,
    ]
    #
    with pytest.raises(UnownedServerShutdown):
        await get(context_b.quit())
    assert context_a.boot_status == BootStatus.ONLINE and context_a.is_owner
    assert context_b.boot_status == BootStatus.ONLINE and not context_b.is_owner
    assert events_a == [
        ServerLifecycleEvent.BOOTING,
        ServerLifecycleEvent.PROCESS_BOOTED,
        ServerLifecycleEvent.CONNECTING,
        ServerLifecycleEvent.OSC_CONNECTED,
        ServerLifecycleEvent.CONNECTED,
        ServerLifecycleEvent.BOOTED,
    ]
    assert events_b == [
        ServerLifecycleEvent.CONNECTING,
        ServerLifecycleEvent.OSC_CONNECTED,
        ServerLifecycleEvent.CONNECTED,
    ]


@pytest.mark.asyncio
@pytest.mark.parametrize("executable", ["scsynth", supernova])
@pytest.mark.parametrize(
    "async_callback, context_class",
    [(False, AsyncServer), (True, AsyncServer), (False, Server)],
)
async def test_boot_a_and_connect_b_and_force_quit_b(
    async_callback, context_class, executable
):
    logger.warning("START")
    context_a, events_a = setup_context(
        context_class, async_callback=async_callback, name="one"
    )
    context_b, events_b = setup_context(
        context_class, async_callback=async_callback, name="two"
    )
    assert context_a.boot_status == BootStatus.OFFLINE and not context_a.is_owner
    assert context_b.boot_status == BootStatus.OFFLINE and not context_b.is_owner
    #
    logger.warning("BOOT A")
    await get(context_a.boot(maximum_logins=2, executable=executable))
    assert events_a == [
        ServerLifecycleEvent.BOOTING,
        ServerLifecycleEvent.PROCESS_BOOTED,
        ServerLifecycleEvent.CONNECTING,
        ServerLifecycleEvent.OSC_CONNECTED,
        ServerLifecycleEvent.CONNECTED,
        ServerLifecycleEvent.BOOTED,
    ]
    assert events_b == []
    #
    logger.warning("CONNECT B")
    await get(context_b.connect())
    assert context_a.boot_status == BootStatus.ONLINE and context_a.is_owner
    assert context_b.boot_status == BootStatus.ONLINE and not context_b.is_owner
    assert events_a == [
        ServerLifecycleEvent.BOOTING,
        ServerLifecycleEvent.PROCESS_BOOTED,
        ServerLifecycleEvent.CONNECTING,
        ServerLifecycleEvent.OSC_CONNECTED,
        ServerLifecycleEvent.CONNECTED,
        ServerLifecycleEvent.BOOTED,
    ]
    assert events_b == [
        ServerLifecycleEvent.CONNECTING,
        ServerLifecycleEvent.OSC_CONNECTED,
        ServerLifecycleEvent.CONNECTED,
    ]
    #
    logger.warning("FORCE PROCESS_QUIT B")
    await get(context_b.quit(force=True))
    assert context_b.boot_status == BootStatus.OFFLINE and not context_b.is_owner
    logger.warning("AWAIT A")
    await get_future(context_a.exit_future)
    assert context_a.boot_status == BootStatus.OFFLINE and not context_a.is_owner
    assert events_a == [
        ServerLifecycleEvent.BOOTING,
        ServerLifecycleEvent.PROCESS_BOOTED,
        ServerLifecycleEvent.CONNECTING,
        ServerLifecycleEvent.OSC_CONNECTED,
        ServerLifecycleEvent.CONNECTED,
        ServerLifecycleEvent.BOOTED,
        ServerLifecycleEvent.PROCESS_PANICKED,
        ServerLifecycleEvent.OSC_PANICKED,
        ServerLifecycleEvent.QUITTING,
        ServerLifecycleEvent.DISCONNECTING,
        ServerLifecycleEvent.DISCONNECTED,
        ServerLifecycleEvent.QUIT,
    ]
    assert events_b == [
        ServerLifecycleEvent.CONNECTING,
        ServerLifecycleEvent.OSC_CONNECTED,
        ServerLifecycleEvent.CONNECTED,
        ServerLifecycleEvent.QUITTING,
        ServerLifecycleEvent.DISCONNECTING,
        ServerLifecycleEvent.OSC_DISCONNECTED,
        ServerLifecycleEvent.DISCONNECTED,
        ServerLifecycleEvent.QUIT,
    ]
    logger.warning("END")


@pytest.mark.asyncio
@pytest.mark.parametrize("executable", ["scsynth", supernova])
@pytest.mark.parametrize("context_class", [AsyncServer, Server])
async def test_boot_reboot_sticky_options(context_class, executable):
    """
    Options persist across booting and quitting.
    """
    context, _ = setup_context(context_class)
    maximum_node_count = random.randint(1024, 2048)
    await get(
        context.boot(maximum_node_count=maximum_node_count, port=find_free_port())
    )
    assert context.options.maximum_node_count == maximum_node_count
    await get(context.quit())
    assert context.options.maximum_node_count == maximum_node_count
    await get(context.boot(port=find_free_port()))
    assert context.options.maximum_node_count == maximum_node_count
    await get(context.quit())
    assert context.options.maximum_node_count == maximum_node_count


@pytest.mark.asyncio
@pytest.mark.parametrize("executable", ["scsynth", supernova])
@pytest.mark.parametrize(
    "async_callback, context_class",
    [(False, AsyncServer), (True, AsyncServer), (False, Server)],
)
async def test_boot_a_and_connect_b_and_kill(
    async_callback, context_class, executable
) -> None:
    logger.warning("START")
    context_a, events_a = setup_context(
        context_class, async_callback=async_callback, name="one"
    )
    context_b, events_b = setup_context(
        context_class, async_callback=async_callback, name="two"
    )
    logger.warning("BOOT A")
    await get(context_a.boot(executable=executable, maximum_logins=2))
    assert events_a == [
        ServerLifecycleEvent.BOOTING,
        ServerLifecycleEvent.PROCESS_BOOTED,
        ServerLifecycleEvent.CONNECTING,
        ServerLifecycleEvent.OSC_CONNECTED,
        ServerLifecycleEvent.CONNECTED,
        ServerLifecycleEvent.BOOTED,
    ]
    assert events_b == []
    logger.warning("CONNECT B")
    await get(context_b.connect())
    assert events_a == [
        ServerLifecycleEvent.BOOTING,
        ServerLifecycleEvent.PROCESS_BOOTED,
        ServerLifecycleEvent.CONNECTING,
        ServerLifecycleEvent.OSC_CONNECTED,
        ServerLifecycleEvent.CONNECTED,
        ServerLifecycleEvent.BOOTED,
    ]
    assert events_b == [
        ServerLifecycleEvent.CONNECTING,
        ServerLifecycleEvent.OSC_CONNECTED,
        ServerLifecycleEvent.CONNECTED,
    ]
    logger.warning("KILL")
    kill()
    logger.warning("AWAIT A AND B")
    await get_future(context_a.exit_future)
    await get_future(context_b.exit_future)
    assert events_a == [
        ServerLifecycleEvent.BOOTING,
        ServerLifecycleEvent.PROCESS_BOOTED,
        ServerLifecycleEvent.CONNECTING,
        ServerLifecycleEvent.OSC_CONNECTED,
        ServerLifecycleEvent.CONNECTED,
        ServerLifecycleEvent.BOOTED,
        ServerLifecycleEvent.PROCESS_PANICKED,
        ServerLifecycleEvent.OSC_PANICKED,
        ServerLifecycleEvent.QUITTING,
        ServerLifecycleEvent.DISCONNECTING,
        ServerLifecycleEvent.DISCONNECTED,
        ServerLifecycleEvent.QUIT,
    ]
    assert events_b == [
        ServerLifecycleEvent.CONNECTING,
        ServerLifecycleEvent.OSC_CONNECTED,
        ServerLifecycleEvent.CONNECTED,
        ServerLifecycleEvent.OSC_PANICKED,
        ServerLifecycleEvent.DISCONNECTING,
        ServerLifecycleEvent.DISCONNECTED,
    ]
    assert context_a.boot_future.done()
    assert context_a.exit_future.done()
    assert context_b.boot_future.done()
    assert context_b.exit_future.done()
    logger.warning("END")



import uuid
from unittest.mock import Mock, call

import pytest

from supriya import AddAction, CalculationRate
from supriya.assets.synthdefs import default
from supriya.contexts import BusGroup, Server, Synth
from supriya.patterns.events import BusAllocateEvent, NoteEvent, Priority

id_ = uuid.uuid4()
bus_id = uuid.uuid4()


@pytest.mark.parametrize(
    "event, offset, expected",
    [
        (
            NoteEvent(id_),
            0.0,
            [
                (0.0, Priority.START, NoteEvent((id_, 0))),
                (1.0, Priority.STOP, NoteEvent((id_, 0))),
            ],
        ),
        (
            NoteEvent(id_, frequency=[440, 550]),
            0.0,
            [
                (0.0, Priority.START, NoteEvent((id_, 0), frequency=440)),
                (0.0, Priority.START, NoteEvent((id_, 1), frequency=550)),
                (1.0, Priority.STOP, NoteEvent((id_, 0), frequency=440)),
                (1.0, Priority.STOP, NoteEvent((id_, 1), frequency=550)),
            ],
        ),
        (
            NoteEvent(id_, frequency=[440, 550], amplitude=[0.5, 0.75, 1.0]),
            2.5,
            [
                (
                    2.5,
                    Priority.START,
                    NoteEvent((id_, 0), amplitude=0.5, frequency=440),
                ),
                (
                    2.5,
                    Priority.START,
                    NoteEvent((id_, 1), amplitude=0.75, frequency=550),
                ),
                (
                    2.5,
                    Priority.START,
                    NoteEvent((id_, 2), amplitude=1.0, frequency=440),
                ),
                (3.5, Priority.STOP, NoteEvent((id_, 0), amplitude=0.5, frequency=440)),
                (
                    3.5,
                    Priority.STOP,
                    NoteEvent((id_, 1), amplitude=0.75, frequency=550),
                ),
                (3.5, Priority.STOP, NoteEvent((id_, 2), amplitude=1.0, frequency=440)),
            ],
        ),
    ],
)
def test_expand(event, offset, expected):
    actual = event.expand(offset)
    assert actual == expected


def test_perform():
    context = Server().boot()
    proxy_mapping = {}
    notes_mapping = {}
    # 0: Allocate a bus for later
    bus_event = BusAllocateEvent(bus_id, calculation_rate="control")
    with context.at():
        bus_event.perform(
            context,
            proxy_mapping,
            current_offset=0.0,
            notes_mapping=notes_mapping,
            priority=Priority.START,
        )
    # A: Allocate
    event_one = NoteEvent(id_)
    spy = Mock(wraps=context)
    with context.at():
        event_one.perform(
            spy,
            proxy_mapping,
            current_offset=0.0,
            notes_mapping=notes_mapping,
            priority=Priority.START,
        )
    assert proxy_mapping == {
        id_: Synth(context=context, id_=1000, synthdef=default),
        bus_id: BusGroup(
            context=context,
            id_=0,
            calculation_rate=CalculationRate.CONTROL,
            count=1,
        ),
    }
    assert notes_mapping == {id_: 1.0}
    assert spy.mock_calls == [
        call.add_synth(
            add_action=AddAction.ADD_TO_HEAD,
            permanent=False,
            synthdef=default,
            target_node=None,
        )
    ]
    # Wait
    context.sync()
    # B: Already allocated, so update settings
    event_two = NoteEvent(id_, amplitude=bus_id, frequency=550)
    spy = Mock(wraps=context)
    with context.at():
        event_two.perform(
            spy,
            proxy_mapping,
            current_offset=1.0,
            notes_mapping=notes_mapping,
            priority=Priority.START,
        )
    assert proxy_mapping == {
        id_: Synth(context=context, id_=1000, synthdef=default),
        bus_id: BusGroup(
            context=context,
            id_=0,
            count=1,
            calculation_rate=CalculationRate.CONTROL,
        ),
    }
    assert notes_mapping == {id_: 2.0}
    assert spy.mock_calls == [
        call.set_node(
            Synth(context=context, id_=1000, synthdef=default),
            amplitude=proxy_mapping[bus_id],
            frequency=550,
        )
    ]
    # C: Free, but stop time doesn't match: no-op
    spy = Mock(wraps=context)
    with context.at():
        event_one.perform(
            spy,
            proxy_mapping,
            current_offset=1.0,
            notes_mapping=notes_mapping,
            priority=Priority.STOP,
        )
    assert proxy_mapping == {
        id_: Synth(context=context, id_=1000, synthdef=default),
        bus_id: BusGroup(
            context=context,
            id_=0,
            calculation_rate=CalculationRate.CONTROL,
            count=1,
        ),
    }
    assert notes_mapping == {id_: 2.0}
    assert spy.mock_calls == []
    # D: Free, and stop time does match
    spy = Mock(wraps=context)
    with context.at():
        event_one.perform(
            spy,
            proxy_mapping,
            current_offset=2.0,
            notes_mapping=notes_mapping,
            priority=Priority.STOP,
        )
    assert proxy_mapping == {
        bus_id: BusGroup(
            context=context,
            id_=0,
            calculation_rate=CalculationRate.CONTROL,
            count=1,
        )
    }
    assert notes_mapping == {}
    assert spy.mock_calls == [
        call.free_node(Synth(context=context, id_=1000, synthdef=default))
    ]



import asyncio

import pytest
import pytest_asyncio

import supriya
from supriya import scsynth
from supriya.contexts.realtime import BaseServer


@pytest.fixture(autouse=True)
def add_libraries(doctest_namespace):
    doctest_namespace["supriya"] = supriya


@pytest.fixture(autouse=True, scope="session")
def shutdown_scsynth():
    scsynth.kill()
    yield
    scsynth.kill()


@pytest_asyncio.fixture(autouse=True)
async def shutdown_realtime_contexts(shutdown_scsynth):
    for context in tuple(BaseServer._contexts):
        result = context._shutdown()
        if asyncio.iscoroutine(result):
            await result
    yield
    for context in tuple(BaseServer._contexts):
        result = context._shutdown()
        if asyncio.iscoroutine(result):
            await result



"""
Core interfaces for interfacting with :term:`scsynth`-compatible execution contexts.

Context subclasses expose a common interface for realtime and non-realtime synthesis.
"""

import abc
import contextlib
import dataclasses
import itertools
import shlex
import threading
from os import PathLike
from typing import (
    Callable,
    Dict,
    List,
    Optional,
    Sequence,
    SupportsFloat,
    SupportsInt,
    Tuple,
    Type,
    Union,
    cast,
)

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal  # type: ignore

from uqbar.objects import new

from ..enums import AddAction, BootStatus, CalculationRate, ParameterRate
from ..scsynth import Options
from ..typing import (
    AddActionLike,
    CalculationRateLike,
    HeaderFormatLike,
    SampleFormatLike,
    SupportsOsc,
)
from ..ugens import SynthDef
from .allocators import BlockAllocator, NodeIdAllocator
from .entities import (
    Buffer,
    BufferGroup,
    Bus,
    BusGroup,
    ContextObject,
    Group,
    Node,
    RootNode,
    Synth,
)
from .errors import AllocationError, ContextError, InvalidCalculationRate, MomentClosed
from .requests import (
    AllocateBuffer,
    AllocateReadBuffer,
    AllocateReadBufferChannel,
    ClearSchedule,
    CloseBuffer,
    CopyBuffer,
    DoNothing,
    FillBuffer,
    FillControlBusRange,
    FreeAllSynthDefs,
    FreeBuffer,
    FreeGroupChildren,
    FreeGroupDeep,
    FreeSynthDef,
    GenerateBuffer,
    LoadSynthDefDirectory,
    LoadSynthDefs,
    MapAudioBusToNode,
    MapControlBusToNode,
    MoveNodeAfter,
    MoveNodeBefore,
    MoveNodeToGroupHead,
    MoveNodeToGroupTail,
    NewGroup,
    NewParallelGroup,
    NewSynth,
    NormalizeBuffer,
    OrderNodes,
    ReadBuffer,
    ReadBufferChannel,
    ReceiveSynthDefs,
    ReleaseNode,
    Request,
    RequestBundle,
    RunNode,
    SetBuffer,
    SetBufferRange,
    SetControlBus,
    SetControlBusRange,
    SetNodeControl,
    SetNodeControlRange,
    WriteBuffer,
    ZeroBuffer,
)


@dataclasses.dataclass
class Moment:
    """
    A context manager representing a moment in time when requests are made to an execution context.

    Multiple requests made inside a moment are bundled together.

    :param context: The moment's context.
    :param seconds: The moment's timestamp.
    """

    context: "Context"
    seconds: Optional[float] = None
    closed: bool = dataclasses.field(default=False, init=False)
    requests: List[Tuple[Request, Optional["Completion"]]] = dataclasses.field(
        default_factory=list, init=False
    )

    def __enter__(self) -> "Moment":
        """
        Set this moment the current "request context".
        """
        if self.closed:
            raise MomentClosed
        self.context._push_moment(self)
        return self

    def __exit__(self, *args) -> None:
        """
        Unset this moment as the current "request context".
        """
        self.context._pop_moment()
        requests = self.context._apply_completions(self.requests)
        timestamp = (
            self.seconds + self.context._latency if self.seconds is not None else None
        )
        if len(requests) and timestamp is not None:
            self.context.send(RequestBundle(timestamp=timestamp, contents=requests))
        elif len(requests) > 1:
            self.context.send(RequestBundle(contents=requests))
        elif len(requests):
            self.context.send(requests[0])
        self.closed = True


@dataclasses.dataclass
class Completion:
    """
    A context manager for collecting requests to be made "on completion" of another request.

    Multiple requests made inside a completion are bundled together.

    :param context: The completion's context.
    :param moment: The completion's moment.
    """

    context: "Context"
    moment: Moment
    requests: List[Tuple[Request, Optional["Completion"]]] = dataclasses.field(
        default_factory=list, init=False
    )

    def __call__(self, request: Request) -> Request:
        """
        Bundle this completion's collected requests into the ``on_completion`` argument of the request.

        :param request: The request to complete.
        """
        if not hasattr(request, "on_completion"):
            raise ValueError(request)
        requests = self.context._apply_completions(self.requests)
        if len(requests) > 1:
            request = new(request, on_completion=RequestBundle(contents=requests))
        elif len(requests) == 1:
            request = new(request, on_completion=requests[0])
        return request

    def __enter__(self) -> "Completion":
        """
        Set this completion as the current "request context".
        """
        if self.moment.closed:
            raise MomentClosed
        self.context._push_completion(self)
        return self

    def __exit__(self, *args) -> None:
        """
        Unset this completion as the current "request context".
        """
        self.context._pop_completion()


class Context(metaclass=abc.ABCMeta):
    """
    A synthesis execution context.

    :param options: The context's options.
    :param kwargs: Keyword arguments for options.
    """

    ### INITIALIZER ###

    def __init__(
        self,
        options: Optional[Options],
        name: Optional[str] = None,
        **kwargs,
    ) -> None:
        self._audio_bus_allocator = BlockAllocator()
        self._boot_status = BootStatus.OFFLINE
        self._buffer_allocator = BlockAllocator()
        self._client_id = 0
        self._control_bus_allocator = BlockAllocator()
        self._latency = 0.0
        self._lock = threading.RLock()
        self._name = name
        self._node_id_allocator = NodeIdAllocator()
        self._options = new(options or Options(), **kwargs)
        self._sync_id = self._sync_id_minimum = 0
        self._sync_id_maximum = 32 << 26
        self._thread_local = threading.local()
        self._setup_allocators()

    ### SPECIAL METHODS ###

    def __getstate__(self):
        state = self.__dict__.copy()
        del state["_lock"]
        del state["_thread_local"]
        return state

    def __repr__(self) -> str:
        return f"<{type(self).__name__} {self.boot_status.name} [{shlex.join(self.options)}]>"

    def __setstate__(self, state):
        self.__dict__.update(state)
        self._lock = threading.RLock()
        self._thread_local = threading.local()

    ### PRIVATE METHODS ###

    def _add_requests(self, *requests: Request) -> None:
        with contextlib.ExitStack() as stack:
            current_requests = (
                self._get_request_context() or stack.enter_context(self.at())
            ).requests
            for request in requests:
                current_requests.append((request, None))

    def _add_request_with_completion(
        self, request: Request, on_completion: Optional[Callable[["Context"], None]]
    ) -> Completion:
        with contextlib.ExitStack() as stack:
            current_requests = (
                self._get_request_context() or stack.enter_context(self.at())
            ).requests
            moment = self._get_moment()
            if moment is None:
                raise ContextError
            completion = Completion(context=self, moment=moment)
            current_requests.append((request, completion))
            if on_completion:
                stack.enter_context(completion)
                on_completion(self)
        return completion

    def _allocate_id(
        self,
        type_: Type[ContextObject],
        calculation_rate: Optional[CalculationRate] = None,
        count: int = 1,
        permanent: bool = False,
    ) -> int:
        id_: Optional[int] = None
        if type_ is Node:
            if permanent:
                id_ = self._node_id_allocator.allocate_permanent_node_id()
            else:
                id_ = self._node_id_allocator.allocate_node_id()
        elif type_ is Buffer:
            id_ = self._buffer_allocator.allocate(count)
        elif type_ is Bus:
            if calculation_rate is CalculationRate.AUDIO:
                id_ = self._audio_bus_allocator.allocate(count)
            elif calculation_rate is CalculationRate.CONTROL:
                id_ = self._control_bus_allocator.allocate(count)
            else:
                raise ValueError(calculation_rate)
        else:
            raise ValueError(type_)
        if id_ is None:
            raise AllocationError
        return id_

    @staticmethod
    def _apply_completions(
        pairs: List[Tuple[Request, Optional[Completion]]]
    ) -> List[Request]:
        requests: List[Request] = []
        for key, group in itertools.groupby(
            [
                request if completion is None else completion(request)
                for request, completion in pairs
            ],
            key=lambda x: type(x),
        ):
            requests.extend(key.merge(list(group)))
        return requests

    @abc.abstractmethod
    def _free_id(
        self,
        type_: Type[ContextObject],
        id_: int,
        calculation_rate: Optional[CalculationRate] = None,
    ) -> None:
        raise NotImplementedError

    def _get_allocator(
        self,
        type_: Type[ContextObject],
        calculation_rate: Optional[CalculationRate] = None,
    ) -> Union[BlockAllocator, NodeIdAllocator]:
        if type_ is Node:
            return self._node_id_allocator
        if type_ is Buffer:
            return self._buffer_allocator
        if type_ is Bus:
            if calculation_rate is CalculationRate.AUDIO:
                return self._audio_bus_allocator
            elif calculation_rate is CalculationRate.CONTROL:
                return self._control_bus_allocator
        raise ValueError

    def _get_moment(self) -> Optional[Moment]:
        moments = self._thread_local.__dict__.get("moments", [])
        if not moments:
            return None
        return moments[-1]

    def _get_next_sync_id(self) -> int:
        with self._lock:
            sync_id = self._sync_id
            self._sync_id += 1
            if self._sync_id > self._sync_id_maximum:
                self._sync_id = self._sync_id_minimum
            return sync_id

    def _get_request_context(self) -> Optional[Union[Completion, Moment]]:
        moments = self._thread_local.__dict__.get("moments", [])
        completions = self._thread_local.__dict__.get("completions", [])
        if completions:
            return completions[-1]
        if moments:
            return moments[-1]
        return None

    def _pop_completion(self) -> None:
        self._thread_local.__dict__.setdefault("completions", []).pop()

    def _pop_moment(self) -> None:
        self._thread_local.__dict__.setdefault("moments", []).pop()

    def _push_completion(self, completion: Completion) -> None:
        self._thread_local.__dict__.setdefault("completions", []).append(completion)

    def _push_moment(self, moment: Moment) -> None:
        self._thread_local.__dict__.setdefault("moments", []).append(moment)

    @abc.abstractmethod
    def _resolve_node(self, node: Union[Node, SupportsInt, None]) -> int:
        raise NotImplementedError

    def _setup_allocators(self) -> None:
        # audio buses
        audio_bus_minimum, audio_bus_maximum = self.options.get_audio_bus_ids(
            self.client_id
        )
        self._audio_bus_allocator = BlockAllocator(
            heap_minimum=audio_bus_minimum, heap_maximum=audio_bus_maximum
        )
        # control buses
        control_bus_minimum, control_bus_maximum = self.options.get_control_bus_ids(
            self.client_id
        )
        self._control_bus_allocator = BlockAllocator(
            heap_minimum=control_bus_minimum, heap_maximum=control_bus_maximum
        )
        # buffers
        buffer_minimum, buffer_maximum = self.options.get_buffer_ids(self.client_id)
        self._buffer_allocator = BlockAllocator(
            heap_minimum=buffer_minimum, heap_maximum=buffer_maximum
        )
        # node IDs
        self._node_id_allocator = NodeIdAllocator(
            initial_node_id=self.options.initial_node_id, client_id=self.client_id
        )
        # sync IDs
        self._sync_id_minimum, self._sync_id_maximum = self.options.get_sync_ids(
            self.client_id
        )
        self._sync_id = self._sync_id_minimum

    @abc.abstractmethod
    def _validate_can_request(self) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    def _validate_moment_timestamp(self, seconds: Optional[float]) -> None:
        raise NotImplementedError

    ### PUBLIC METHODS ###

    def add_buffer(
        self,
        *,
        channel_count: Optional[int] = None,
        channel_indices: Optional[List[int]] = None,
        file_path: Optional[PathLike] = None,
        frame_count: Optional[int] = None,
        starting_frame: Optional[int] = None,
        on_completion: Optional[Callable[["Context"], None]] = None,
    ) -> Buffer:
        """
        Add a new buffer to the context.

        Emit ``/b_alloc``, ``/b_allocRead`` or ``/b_allocReadChannel`` requests
        depending on parameters.

        :param channel_count: The channel count of the new buffer. Cannot be used when
            reading from file paths.
        :param channel_indices: The channels to read from a file when reading from a
            file.
        :param file_path: The (optional) file to read from.
        :param frame_count: The frame count of the new buffer.
        :param starting_frame: The frame to start reading from when reading from a file.
        :param on_completion: A callable with the buffer's context as the only argument.
            Permits building an "on completion" argument to this method's request
            without an active moment.
        """
        self._validate_can_request()
        if not (frame_count or file_path):
            raise ValueError
        if channel_count and channel_indices:
            raise ValueError
        if channel_count and file_path:
            raise ValueError
        id_ = self._allocate_id(Buffer)
        if file_path and channel_indices:
            request: Request = AllocateReadBufferChannel(
                buffer_id=id_,
                channel_indices=channel_indices,
                path=file_path,
                frame_count=frame_count or 0,
                starting_frame=starting_frame or 0,
            )
        elif file_path:
            request = AllocateReadBuffer(
                buffer_id=id_,
                path=file_path,
                frame_count=frame_count or 0,
                starting_frame=starting_frame or 0,
            )
        elif frame_count:
            request = AllocateBuffer(
                buffer_id=id_, channel_count=channel_count or 1, frame_count=frame_count
            )
        else:
            raise ValueError
        completion = self._add_request_with_completion(request, on_completion)
        return Buffer(context=self, id_=id_, completion=completion)

    def add_buffer_group(
        self,
        *,
        channel_count: Optional[int] = None,
        count: int = 1,
        frame_count: Optional[int] = None,
    ) -> BufferGroup:
        """
        Add a group of new buffers to the context.

        Emit ``/b_alloc`` requests.

        :param channel_count: The channel count of the new buffers.
        :param count: The number of buffers to add.
        :param frame_count: The frame count of the new buffers.
        """
        self._validate_can_request()
        if not (channel_count and frame_count):
            raise ValueError
        if count < 1:
            raise ValueError
        id_ = self._allocate_id(Buffer, count=count)
        requests: List[Request] = []
        for i in range(count):
            requests.append(
                AllocateBuffer(
                    buffer_id=id_ + i,
                    channel_count=channel_count,
                    frame_count=frame_count,
                )
            )
        self._add_requests(*requests)
        return BufferGroup(
            context=self,
            id_=id_,
            count=count,
        )

    def add_bus(
        self, calculation_rate: CalculationRateLike = CalculationRate.CONTROL
    ) -> Bus:
        """
        Add a new bus to the context.

        Emit no requests.

        :param calculation_rate: The calculation rate of the new bus.
        """
        self._validate_can_request()
        rate = CalculationRate.from_expr(calculation_rate)
        if rate not in (CalculationRate.AUDIO, CalculationRate.CONTROL):
            raise InvalidCalculationRate(rate)
        id_ = self._allocate_id(Bus, calculation_rate=rate)
        return Bus(calculation_rate=rate, context=self, id_=id_)

    def add_bus_group(
        self,
        calculation_rate: CalculationRateLike = CalculationRate.CONTROL,
        count: int = 1,
    ) -> BusGroup:
        """
        Add a group of new buses to the context.

        Emit no requests.

        :param calculation_rate: The calculation rate of the new bus.
        :param count: The number of buses to add.
        """
        self._validate_can_request()
        rate = CalculationRate.from_expr(calculation_rate)
        if rate not in (CalculationRate.AUDIO, CalculationRate.CONTROL):
            raise InvalidCalculationRate(rate)
        if count < 1:
            raise ValueError
        id_ = self._allocate_id(Bus, calculation_rate=rate, count=count)
        return BusGroup(
            calculation_rate=rate,
            context=self,
            count=count,
            id_=id_,
        )

    def add_group(
        self,
        *,
        add_action: AddActionLike = AddAction.ADD_TO_HEAD,
        target_node: Optional[SupportsInt] = None,
        parallel: bool = False,
        permanent: bool = False,
    ) -> Group:
        """
        Add a new group node to the context.

        Emit ``/g_new`` or ``/p_new`` requests depending on parameters.

        :param add_action: The :term:`add action` to use when placing the new group.
        :param target_node: The node to place the new group relative to.
        :param parallel: Flag for parallel vs non-parallel groups.
        :param permanent: Flag for using a permanent node ID.
        """
        self._validate_can_request()
        add_action_ = AddAction.from_expr(add_action)
        if isinstance(target_node, Node):
            if add_action_ not in target_node._valid_add_actions:
                raise ValueError(add_action_)
        target_node_id = self._resolve_node(target_node)
        id_ = self._allocate_id(Node, permanent=permanent)
        items = [(id_, add_action_, target_node_id)]
        if parallel:
            request: Request = NewParallelGroup(items=items)
        else:
            request = NewGroup(items=items)
        self._add_requests(request)
        return Group(context=self, id_=id_, parallel=parallel)

    def add_synth(
        self,
        synthdef: SynthDef,
        *,
        add_action: AddActionLike = AddAction.ADD_TO_HEAD,
        target_node: Optional[SupportsInt] = None,
        permanent: bool = False,
        **settings,
    ) -> Synth:
        """
        Add a new synth node to the context.

        Emit ``/s_new`` requests.

        :param synthdef: The :term:`SynthDef` to use for the new synth.
        :param add_action: The :term:`add action` to use when placing the new synth.
        :param target_node: The node to place the new synth relative to.
        :param permanent: Flag for using a permanent node ID.
        :param settings: The new synth's control settings.
        """
        self._validate_can_request()
        add_action_ = AddAction.from_expr(add_action)
        if isinstance(target_node, Node):
            if add_action_ not in target_node._valid_add_actions:
                raise ValueError(add_action_)
        target_node_id = self._resolve_node(target_node)
        synthdef_kwargs: Dict[
            Union[int, str], Union[float, str, Tuple[Union[float, str], ...]]
        ] = {}
        for _, parameter in synthdef.indexed_parameters:
            if parameter.name not in settings:
                continue
            value = settings[parameter.name]
            if not isinstance(value, Sequence) or isinstance(value, str):
                value = (value,)
            if value == parameter.value:
                continue
            if parameter.rate is ParameterRate.SCALAR or parameter.name in (
                "in_",
                "out",
            ):
                synthdef_kwargs[parameter.name] = tuple(float(v) for v in value)
            else:
                processed_values: List[Union[float, str]] = []
                for v in value:
                    if isinstance(v, Bus):
                        processed_values.append(v.map_symbol())
                    elif isinstance(v, str):
                        processed_values.append(v)
                    else:
                        processed_values.append(float(v))
                if len(processed_values) == 1:
                    synthdef_kwargs[parameter.name] = processed_values[0]
                else:
                    synthdef_kwargs[parameter.name] = tuple(processed_values)
        id_ = self._allocate_id(Node, permanent=permanent)
        self._add_requests(
            NewSynth(
                add_action=add_action_,
                synth_id=id_,
                synthdef=synthdef,
                target_node_id=target_node_id,
                controls=synthdef_kwargs,
            )
        )
        return Synth(context=self, id_=id_, synthdef=synthdef)

    def add_synthdefs(
        self,
        *synthdefs: SynthDef,
        on_completion: Optional[Callable[["Context"], None]] = None,
    ) -> Completion:
        """
        Add one or more SynthDefs to the context.

        Emit ``/d_recv`` requests.

        :param synthdefs: The synthdefs to add.
        :param on_completion: A callable with the buffer's context as the only argument.
            Permits building an "on completion" argument to this method's request
            without an active moment.
        """
        self._validate_can_request()
        if not synthdefs:
            raise ValueError
        request = ReceiveSynthDefs(synthdefs=synthdefs)
        return self._add_request_with_completion(request, on_completion)

    def at(self, seconds=None) -> Moment:
        """
        Create a Moment.

        :param seconds: The timestamp of the new moment.
        """
        self._validate_moment_timestamp(seconds)
        return Moment(context=self, seconds=seconds)

    def clear_schedule(self) -> None:
        """
        Clear all scheduled bundles.

        Emit ``/clearSched`` requests.
        """
        self._validate_can_request()
        request = ClearSchedule()
        self._add_requests(request)

    def close_buffer(
        self,
        buffer: Buffer,
        on_completion: Optional[Callable[["Context"], None]] = None,
    ) -> Completion:
        """
        Close a buffer.

        Emit ``/b_close`` requests.

        :param buffer: The buffer to close.
        :param on_completion: A callable with the buffer's context as the only argument.
            Permits building an "on completion" argument to this method's request
            without an active moment.
        """
        self._validate_can_request()
        request = CloseBuffer(buffer_id=buffer.id_)
        return self._add_request_with_completion(request, on_completion)

    def copy_buffer(
        self,
        *,
        source_buffer: Buffer,
        target_buffer: Buffer,
        source_starting_frame: int,
        target_starting_frame: int,
        frame_count: int,
    ) -> None:
        """
        Copy a buffer.

        Emit ``/b_gen <buffer.id_> copy ...`` requests.

        :param source_buffer: The buffer to copy from.
        :param target_buffer: The buffer to copy to.
        :param source_starting_frame: The frame index in the source buffer to start
            reading from.
        :param target_starting_frame: The frame index in the target buffer to start
            writing at.
        :param frame_count: The number of frames to copy.
        """
        self._validate_can_request()
        request = CopyBuffer(
            frame_count=frame_count,
            source_buffer_id=source_buffer,
            source_starting_frame=source_starting_frame,
            target_buffer_id=target_buffer,
            target_starting_frame=target_starting_frame,
        )
        self._add_requests(request)

    def do_nothing(self) -> None:
        """
        Emit a no-op "nothing" command.
        """
        self._validate_can_request()
        self._add_requests(DoNothing())

    def fill_buffer(
        self, buffer: Buffer, starting_frame: int, frame_count: int, value: float
    ) -> None:
        """
        Fill a buffer with a single value.

        Emit ``/b_fill`` requests.

        :param buffer: The buffer to fill.
        :param starting_frame: The frame index to start filling at.
        :param frame_count: The number of frames to fill.
        :param value: The value to fill with.
        """
        self._validate_can_request()
        request = FillBuffer(
            buffer_id=buffer, items=[(starting_frame, frame_count, value)]
        )
        self._add_requests(request)

    def fill_bus_range(self, bus: Bus, count: int, value: float) -> None:
        """
        Fill a contiguous range of buses with a single value.

        Emit ``/c_fill`` requests.

        :param count: The number of buses to fill.
        :param value: The value to fill with.
        """
        self._validate_can_request()
        if bus.calculation_rate != CalculationRate.CONTROL:
            raise InvalidCalculationRate
        request = FillControlBusRange(items=[(bus.id_, count, value)])
        self._add_requests(request)

    def free_buffer(
        self,
        buffer: Buffer,
        on_completion: Optional[Callable[["Context"], None]] = None,
    ) -> Completion:
        """
        Free a buffer.

        Emit ``/b_free`` requests.

        .. note::

            Freeing the first buffer of a buffer group will free the IDs of all buffers
            in the group, but will only issue a ``/b_free`` request for the first
            buffer.

        :param buffer: The buffer to free.
        :param on_completion: A callable with the buffer's context as the only argument.
            Permits building an "on completion" argument to this method's request
            without an active moment.
        """
        self._validate_can_request()
        request = FreeBuffer(buffer_id=buffer)
        return self._add_request_with_completion(request, on_completion)

    def free_buffer_group(self, buffer_group: BufferGroup) -> None:
        """
        Free a buffer grop.

        Emit ``/b_free`` requests.

        :param buffer_group: The buffer group to free.
        """
        self._validate_can_request()
        with contextlib.ExitStack() as stack:
            if not self._get_request_context():
                stack.enter_context(self.at())
            for buffer_ in buffer_group.buffers:
                self._add_requests(FreeBuffer(buffer_id=buffer_.id_))

    def free_bus(self, bus: Bus) -> None:
        """
        Free a bus.

        Emit no requests.

        .. note::

            Freeing the first bus of a bus group will free the IDs of the entire group.

        :param bus: The bus to free.
        """
        self._validate_can_request()
        self._free_id(Bus, bus.id_, calculation_rate=bus.calculation_rate)

    def free_bus_group(self, bus_group: BusGroup) -> None:
        """
        Free a bus group.

        Emit no requests.

        :param bus_group: The bus group to free.
        """
        self._validate_can_request()
        self._free_id(Bus, bus_group.id_, calculation_rate=bus_group.calculation_rate)

    def free_group_children(self, group: Group, synths_only: bool = False) -> None:
        """
        Free a group's children.

        Emit ``/g_deepFree`` or ``/g_freeAll`` requests depending on parameters.

        :param group: The group whose children will be freed.
        :param synths_only: Flag for freeing only child synths, or all children.
        """
        self._validate_can_request()
        if synths_only:
            request: Request = FreeGroupDeep(node_ids=[group.id_])
        else:
            request = FreeGroupChildren(node_ids=[group.id_])
        self._add_requests(request)

    def free_node(self, node: Node, force: bool = False) -> None:
        """
        Free a node.

        Emit ``/n_free`` for groups, for synths without a ``gate`` control, or when
        ``force`` is ``True``.

        Emit ``/n_set <node.id_> gate 0`` for synths with ``gate`` controls.

        :param node: The node to free.
        :param force: Flag for force-freeing, without releasing.
        """
        self._validate_can_request()
        request = ReleaseNode(
            node.id_,
            force=force,
            has_gate=isinstance(node, Synth) and "gate" in node.synthdef.parameters,
        )
        self._add_requests(request)

    def free_synthdefs(self, *synthdefs: SynthDef) -> None:
        """
        Free one or more SynthDefs.

        Emit ``/d_free`` requests.

        :param synthdefs: The synthdefs to free.
        """
        self._validate_can_request()
        if not synthdefs:
            raise ValueError
        request = FreeSynthDef(synthdefs=synthdefs)
        self._add_requests(request)

    def free_all_synthdefs(self) -> None:
        """
        Free all SynthDefs.

        Emit ``/d_freeAll`` requests.
        """
        self._validate_can_request()
        request = FreeAllSynthDefs()
        self._add_requests(request)

    def generate_buffer(
        self,
        buffer: Buffer,
        command_name: Literal["sine1", "sine2", "sine3", "cheby"],
        amplitudes: Sequence[float],
        frequencies: Optional[Sequence[float]] = None,
        phases: Optional[Sequence[float]] = None,
        as_wavetable: bool = False,
        should_clear_first: bool = False,
        should_normalize: bool = False,
    ) -> None:
        """
        Generate a buffer.

        Emit ``/b_gen`` requests.

        :param buffer: The buffer to generate.
        :param command_name: The generation command name.
        :param amplitudes: A sequence of partial amplitudes.
        :param frequencies: A sequence of partial frequencies.
        :param phases: A sequence of partial phases.
        :param as_wavetable: Flag for generating the output in wavetable format.
        :param should_clear_first: Flag for clearing the buffer before generating.
        :param should_normalize: Flag for normalizing the generated output.
        """
        self._validate_can_request()
        if not amplitudes:
            raise ValueError
        if command_name == "sine2":
            if not frequencies:
                raise ValueError
            elif not (len(amplitudes) == len(frequencies)):
                raise ValueError
        elif command_name == "sine3":
            if not frequencies or not phases:
                raise ValueError
            elif not (len(amplitudes) == len(frequencies) == len(phases)):
                raise ValueError
        request = GenerateBuffer(
            buffer_id=buffer,
            command_name=command_name,
            amplitudes=amplitudes,
            frequencies=frequencies,
            phases=phases,
            as_wavetable=as_wavetable,
            should_clear_first=should_clear_first,
            should_normalize=should_normalize,
        )
        self._add_requests(request)

    def load_synthdefs(
        self,
        path: PathLike,
        on_completion: Optional[Callable[["Context"], None]] = None,
    ) -> Completion:
        """
        Load SynthDefs from a path.

        Emit ``/d_load`` requests.

        .. warning::

            At the time of this writing, supernova does not support globbing.

        :param path: The file path to load from. Globbing characters (e.g. ``*``) are
            permitted.
        :param on_completion: A callable with the buffer's context as the only argument.
            Permits building an "on completion" argument to this method's request
            without an active moment.
        """
        self._validate_can_request()
        request = LoadSynthDefs(path=path)
        return self._add_request_with_completion(request, on_completion)

    def load_synthdefs_directory(
        self,
        path: PathLike,
        on_completion: Optional[Callable[["Context"], None]] = None,
    ) -> Completion:
        """
        Load all SynthDefs from a directory.

        Emit ``/d_loadDir`` requests.

        :param path: The directory path to load from.
        :param on_completion: A callable with the buffer's context as the only argument.
            Permits building an "on completion" argument to this method's request
            without an active moment.
        """
        self._validate_can_request()
        request = LoadSynthDefDirectory(path=path)
        return self._add_request_with_completion(request, on_completion)

    def map_node(self, node: Node, **settings: Union[Bus, None]) -> None:
        """
        Map a node's controls to buses.

        Emit ``/n_map`` and ``/n_mapa`` requests.

        :param node: The node whose controls will be mapped.
        :param settings: A mapping of control names to buses (or to ``None`` to unmap
            the control).
        """
        self._validate_can_request()
        control, audio = {}, {}
        for key, value in settings.items():
            if isinstance(value, Bus):
                if value.calculation_rate is CalculationRate.AUDIO:
                    audio[key] = int(value)
                else:
                    control[key] = int(value)
            elif value is None:
                control[key] = -1
        requests: List[Request] = []
        if control:
            requests.append(
                MapControlBusToNode(node_id=node, items=sorted(control.items()))
            )
        if audio:
            requests.append(
                MapAudioBusToNode(node_id=node, items=sorted(audio.items()))
            )
        self._add_requests(*requests)

    def move_node(
        self, node: Node, add_action: AddActionLike, target_node: Node
    ) -> None:
        """
        Move a node.

        Emit ``/n_after``, ``/n_before``, ``/g_head`` and ``/g_tail`` requests depending
        on parameters.

        :param node: The node to move.
        :param add_action: The :term:`add action` to use when moving the node.
        :param target_node: The target node to place the node relative to.
        """
        self._validate_can_request()
        add_action_ = AddAction.from_expr(add_action)
        items = [(node, target_node)]
        if add_action_ is AddAction.ADD_BEFORE:
            request: Request = MoveNodeBefore(items)
        elif add_action_ is AddAction.ADD_AFTER:
            request = MoveNodeAfter(items)
        elif add_action_ is AddAction.ADD_TO_TAIL:
            request = MoveNodeToGroupTail(items)
        elif add_action_ is AddAction.ADD_TO_HEAD:
            request = MoveNodeToGroupHead(items)
        else:
            raise ValueError
        self._add_requests(request)

    def normalize_buffer(
        self, buffer: Buffer, new_maximum: float = 1.0, as_wavetable: bool = False
    ) -> None:
        """
        Normalize a buffer.

        Emit ``/b_gen <buffer.id_> (w)?normalize`` requests depending on parameters.

        :param buffer: The buffer to normalize.
        :param new_maximum: The new maximum to normalize to.
        :param as_wavetable: Flag for treating the buffer contents as a wavetable.
        """
        self._validate_can_request()
        request = NormalizeBuffer(
            buffer_id=buffer.id_, new_maximum=new_maximum, as_wavetable=as_wavetable
        )
        self._add_requests(request)

    def order_nodes(
        self, target_node: Node, *nodes: Node, add_action: AddActionLike = None
    ) -> None:
        """
        Re-order nodes.

        Emit ``/n_order`` requests.

        :param target_node: The node to re-order the other nodes relative to.
        :param nodes: The nodes to re-order.
        :param add_action: The :term:`add action` to use when re-ordering the nodes.
        """
        self._validate_can_request()
        request = OrderNodes(
            add_action=add_action, target_node_id=target_node, node_ids=nodes
        )
        self._add_requests(request)

    def pause_node(self, node: Node) -> None:
        """
        Pause a node.

        Emit ``/n_run <node.id_> 0`` requests.

        :param node: The node to pause.
        """
        self._validate_can_request()
        request = RunNode(items=[(node, False)])
        self._add_requests(request)

    def read_buffer(
        self,
        buffer: Buffer,
        file_path: PathLike,
        *,
        buffer_starting_frame: Optional[int] = None,
        channel_indices: Optional[List[int]] = None,
        frame_count: Optional[int] = None,
        leave_open: bool = False,
        starting_frame: Optional[int] = None,
        on_completion: Optional[Callable[["Context"], None]] = None,
    ) -> Completion:
        """
        Read a file into a buffer.

        Emit ``/b_read`` or ``/b_readChannel`` requests, depending on parameters.

        :param buffer: The buffer to read into.
        :param file_path: The file path to read from.
        :param channel_indices: A list of channel indices to read from when reading from
            a file.
        :param frame_count: The number of frames to read.
        :param leave_open: Flag for leaving the file open (e.g. to continue reading via
            :py:class:`~supriya.ugens.diskio.DiskIn`) or close it.
        :param starting_frame: The starting frame in the buffer to begin reading into
            at.
        :param on_completion: A callable with the buffer's context as the only argument.
            Permits building an "on completion" argument to this method's request
            without an active moment.
        """
        self._validate_can_request()
        frame_count_ = frame_count or 0
        if channel_indices:
            request: Request = ReadBufferChannel(
                buffer_id=buffer.id_,
                path=file_path,
                frame_count=frame_count_ or -1,
                leave_open=leave_open,
                starting_frame_in_buffer=buffer_starting_frame or 0,
                starting_frame_in_file=starting_frame or 0,
                channel_indices=channel_indices,
            )
        else:
            request = ReadBuffer(
                buffer_id=buffer.id_,
                path=file_path,
                frame_count=frame_count_ or -1,
                leave_open=leave_open,
                starting_frame_in_buffer=buffer_starting_frame or 0,
                starting_frame_in_file=starting_frame or 0,
            )
        return self._add_request_with_completion(request, on_completion)

    @abc.abstractmethod
    def send(self, message: SupportsOsc):
        """
        Send a message to the execution context.

        :param message: The message to send.
        """
        raise NotImplementedError

    def set_buffer(self, buffer: Buffer, index: int, value: float) -> None:
        """
        Set a buffer sample.

        Emit ``/b_set`` requests.

        :param buffer: The buffer to modify.
        :param index: The sample index to write at.
        :param value: The value to write.
        """
        self._validate_can_request()
        request = SetBuffer(buffer_id=buffer, items=[(index, value)])
        self._add_requests(request)

    def set_buffer_range(
        self, buffer: Buffer, index: int, values: Sequence[float]
    ) -> None:
        """
        Set a buffer sample range.

        Emit ``/b_setn`` requests.

        :param buffer: The buffer to modify.
        :param index: The sample index to start writing at.
        :param values: The values to write.
        """
        self._validate_can_request()
        request = SetBufferRange(buffer_id=buffer, items=[(index, values)])
        self._add_requests(request)

    def set_bus(self, bus: Bus, value: float) -> None:
        """
        Set a control bus to a value.

        Emit ``/c_set`` requests.

        :param bus: The control bus to set.
        :param value: The value to set the control bus to.
        """
        self._validate_can_request()
        if bus.calculation_rate != CalculationRate.CONTROL:
            raise InvalidCalculationRate
        request = SetControlBus(items=[(bus.id_, value)])
        self._add_requests(request)

    def set_bus_range(self, bus: Bus, values: Sequence[float]) -> None:
        """
        Set a range of control buses.

        Emit ``/c_setn`` requests.

        :param bus: The bus to start writing at.
        :param values: The values to write.
        """
        self._validate_can_request()
        if bus.calculation_rate != CalculationRate.CONTROL:
            raise InvalidCalculationRate
        request = SetControlBusRange(items=[(bus.id_, values)])
        self._add_requests(request)

    def set_node(
        self,
        node: Node,
        *indexed_settings: Tuple[int, Union[SupportsFloat, Sequence[SupportsFloat]]],
        **settings: Union[SupportsFloat, Sequence[SupportsFloat]],
    ) -> None:
        """
        Set a node's controls.

        Emit ``/n_set`` requests.

        :param node: The node whose controls will be set.
        :param indexed_settings: A sequence of control indices to values.
        :param settings: A mapping of control names to values.
        """
        self._validate_can_request()
        coerced_settings: Dict[Union[int, str], Union[float, Sequence[float]]] = {}
        for index, values in sorted(indexed_settings):
            if isinstance(values, Sequence):
                coerced_settings[index] = [float(value) for value in values]
            else:
                coerced_settings[index] = float(values)
        for key, values in sorted(settings.items()):
            if isinstance(values, Sequence):
                coerced_settings[key] = [float(value) for value in values]
            else:
                coerced_settings[key] = float(values)
        request = SetNodeControl(node_id=node.id_, items=list(coerced_settings.items()))
        self._add_requests(request)

    def set_node_range(
        self,
        node: Node,
        *indexed_settings: Tuple[int, Sequence[SupportsFloat]],
        **settings: Sequence[SupportsFloat],
    ) -> None:
        """
        Set a range of node controls.

        Emit ``/n_setn`` requests.

        :param node: The node whose controls will be set.
        :param indexed_settings: A sequence of control indices to values.
        :param settings: A mapping of control names to values.
        """
        self._validate_can_request()
        coerced_settings: Dict[Union[int, str], Sequence[float]] = {}
        for index, values in sorted(indexed_settings):
            coerced_settings[index] = [float(value) for value in values]
        for key, values in sorted(settings.items()):
            coerced_settings[key] = [float(value) for value in values]
        request = SetNodeControlRange(
            node_id=node.id_, items=list(coerced_settings.items())
        )
        self._add_requests(request)

    def unpause_node(self, node: Node) -> None:
        """
        Unpause a node.

        Emit ``/n_run <node.id_> 1`` requests.

        :param node: The node to unpause.
        """
        self._validate_can_request()
        request = RunNode(items=[(node, True)])
        self._add_requests(request)

    def write_buffer(
        self,
        buffer: Buffer,
        file_path: PathLike,
        *,
        frame_count: Optional[int] = None,
        header_format: HeaderFormatLike = "aiff",
        leave_open: bool = False,
        sample_format: SampleFormatLike = "int24",
        starting_frame: Optional[int] = None,
        on_completion: Optional[Callable[["Context"], None]] = None,
    ) -> Completion:
        """
        Write a buffer to disk.

        Emit ``/b_write`` requests.

        :param buffer: The buffer to write to disk.
        :param file_path: The file path to write into.
        :param frame_count: The number of frames to write.
        :param header_format: The header format to use, e.g. ``AIFF`` or ``WAVE``.
        :param leave_open: Flag for leaving the file open (e.g. to continue writing via
            :py:class:`~supriya.ugens.diskio.DiskOut`) or close it.
        :param sample_format: The sample format to use, e.g. ``INT24`` or ``FLOAT``.
        :param starting_frame: The starting frame in the buffer to start writing from.
        :param on_completion: A callable with the buffer's context as the only argument.
            Permits building an "on completion" argument to this method's request
            without an active moment.
        """
        self._validate_can_request()
        request = WriteBuffer(
            buffer_id=buffer.id_,
            path=file_path,
            frame_count=frame_count or -1,
            header_format=header_format,
            leave_open=leave_open,
            sample_format=sample_format,
            starting_frame=starting_frame or 0,
        )
        return self._add_request_with_completion(request, on_completion)

    def zero_buffer(
        self,
        buffer: Buffer,
        on_completion: Optional[Callable[["Context"], None]] = None,
    ) -> Completion:
        """
        Set a buffer's contents to zero.

        Emit ``/b_zero`` requests.

        :param buffer: The buffer to zero.
        :param on_completion: A callable with the buffer's context as the only argument.
            Permits building an "on completion" argument to this method's request
            without an active moment.
        """
        self._validate_can_request()
        request = ZeroBuffer(buffer_id=buffer)
        return self._add_request_with_completion(request, on_completion)

    ### PUBLIC PROPERTIES ###

    @property
    def audio_input_bus_group(self) -> BusGroup:
        """
        Get the context's audio output buses.
        """
        return BusGroup(
            context=self,
            id_=self.options.output_bus_channel_count,
            calculation_rate=cast(CalculationRate, CalculationRate.AUDIO),
            count=self.options.input_bus_channel_count,
        )

    @property
    def audio_output_bus_group(self) -> BusGroup:
        """
        Get the context's audio input buses.
        """
        return BusGroup(
            context=self,
            id_=0,
            calculation_rate=cast(CalculationRate, CalculationRate.AUDIO),
            count=self.options.output_bus_channel_count,
        )

    @property
    def boot_status(self) -> BootStatus:
        """
        Get the server's boot status.
        """
        return self._boot_status

    @property
    def default_group(self) -> Group:
        """
        Get the server's default group.
        """
        return self.root_node

    @property
    def client_id(self) -> int:
        """
        Get the context's client ID.
        """
        return self._client_id

    @property
    def latency(self) -> float:
        """
        Get the context's latency.
        """
        return self._latency

    @property
    def name(self) -> Optional[str]:
        """
        Get the context's optional name.
        """
        return self._name

    @property
    def options(self) -> Options:
        """
        Get the context's scsynth options.
        """
        return self._options

    @property
    def root_node(self) -> RootNode:
        """
        Get the context's root node.
        """
        return RootNode(context=self, id_=0)



import subprocess
from pathlib import Path

import uqbar.io


def websafe_audio(output_path: Path) -> Path:
    # Convert to MP3 if possible for smaller file sizes:
    if uqbar.io.find_executable("lame"):
        new_output_path = output_path.with_suffix(".mp3")
        if new_output_path.exists():
            return new_output_path
        command = "lame -V2 {} {}".format(output_path, new_output_path)
        exit_code = subprocess.call(
            command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )
        if not exit_code:
            return new_output_path
    # If MP3-conversion, fails, try to convert to .wav
    if uqbar.io.find_executable("ffmpeg"):
        new_output_path = output_path.with_suffix(".wav")
        command = "ffmpeg -y -i {} {}".format(output_path, new_output_path)
        exit_code = subprocess.call(
            command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )
        if not exit_code:
            return new_output_path
    return output_path



from .core import UGen, param, ugen


@ugen(ir=True)
class BlockSize(UGen):
    """
    A block size info unit generator.

    ::

        >>> supriya.ugens.BlockSize.ir()
        <BlockSize.ir()[0]>
    """


@ugen(kr=True, ir=True)
class BufChannels(UGen):
    """
    A buffer channel count info unit generator.

    ::

        >>> supriya.ugens.BufChannels.kr(buffer_id=0)
        <BufChannels.kr()[0]>
    """

    buffer_id = param()


@ugen(kr=True, ir=True)
class BufDur(UGen):
    """
    A buffer duration info unit generator.

    ::

        >>> supriya.ugens.BufDur.kr(buffer_id=0)
        <BufDur.kr()[0]>
    """

    buffer_id = param()


@ugen(kr=True, ir=True)
class BufFrames(UGen):
    """
    A buffer frame count info unit generator.

    ::

        >>> supriya.ugens.BufFrames.kr(buffer_id=0)
        <BufFrames.kr()[0]>
    """

    buffer_id = param()


@ugen(kr=True, ir=True)
class BufRateScale(UGen):
    """
    A buffer sample-rate scale info unit generator.

    ::

        >>> supriya.ugens.BufRateScale.kr(buffer_id=0)
        <BufRateScale.kr()[0]>
    """

    buffer_id = param()


@ugen(kr=True, ir=True)
class BufSampleRate(UGen):
    """
    A buffer sample-rate info unit generator.

    ::

        >>> supriya.ugens.BufSampleRate.kr(buffer_id=0)
        <BufSampleRate.kr()[0]>
    """

    buffer_id = param()


@ugen(kr=True, ir=True)
class BufSamples(UGen):
    """
    A buffer sample count info unit generator.

    ::

        >>> supriya.ugens.BufSamples.kr(buffer_id=0)
        <BufSamples.kr()[0]>
    """

    buffer_id = param()


@ugen(ir=True)
class ControlDur(UGen):
    """
    A control duration info unit generator.

    ::

        >>> supriya.ugens.ControlDur.ir()
        <ControlDur.ir()[0]>
    """


@ugen(ir=True)
class ControlRate(UGen):
    """
    A control rate info unit generator.

    ::

        >>> supriya.ugens.ControlRate.ir()
        <ControlRate.ir()[0]>
    """


@ugen(ir=True)
class NodeID(UGen):
    """
    A node ID info unit generator.

    ::

        >>> supriya.ugens.NodeID.ir()
        <NodeID.ir()[0]>
    """


@ugen(ir=True)
class NumAudioBuses(UGen):
    """
    A number of audio buses info unit generator.

    ::

        >>> supriya.ugens.NumAudioBuses.ir()
        <NumAudioBuses.ir()[0]>
    """


@ugen(ir=True)
class NumBuffers(UGen):
    """
    A number of buffers info unit generator.

    ::

        >>> supriya.ugens.NumBuffers.ir()
        <NumBuffers.ir()[0]>
    """


@ugen(ir=True)
class NumControlBuses(UGen):
    """
    A number of control buses info unit generator.

    ::

        >>> supriya.ugens.NumControlBuses.ir()
        <NumControlBuses.ir()[0]>
    """


@ugen(ir=True)
class NumInputBuses(UGen):
    """
    A number of input buses info unit generator.

    ::

        >>> supriya.ugens.NumInputBuses.ir()
        <NumInputBuses.ir()[0]>
    """


@ugen(ir=True)
class NumOutputBuses(UGen):
    """
    A number of output buses info unit generator.

    ::

        >>> supriya.ugens.NumOutputBuses.ir()
        <NumOutputBuses.ir()[0]>
    """


@ugen(kr=True, ir=True)
class NumRunningSynths(UGen):
    """
    A number of running synths info unit generator.

    ::

        >>> supriya.ugens.NumRunningSynths.ir()
        <NumRunningSynths.ir()[0]>
    """


@ugen(ir=True)
class RadiansPerSample(UGen):
    """
    A radians-per-sample info unit generator.

    ::

        >>> supriya.ugens.RadiansPerSample.ir()
        <RadiansPerSample.ir()[0]>
    """


@ugen(ir=True)
class SampleDur(UGen):
    """
    A sample duration info unit generator.

    ::

        >>> supriya.ugens.SampleDur.ir()
        <SampleDur.ir()[0]>
    """


@ugen(ir=True)
class SampleRate(UGen):
    """
    A sample-rate info unit generator.

    ::

        >>> supriya.ugens.SampleRate.ir()
        <SampleRate.ir()[0]>
    """


@ugen(ir=True)
class SubsampleOffset(UGen):
    """
    A subsample-offset info unit generator.

    ::

        >>> supriya.ugens.SubsampleOffset.ir()
        <SubsampleOffset.ir()[0]>
    """
